/* lex2html -- convert lexicon to html for viewing */

/* Mark Huckvale - University College London */

/* version 1.0 - March 1998 */

#define PROGNAME "lex2html"
#define PROGVERS "1.0"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "lt-memory.h"
#include "nsl.h"
#include "cgi.h"

char	basedir[256];		/* WBT base directory */
char	cgibase[256];		/* base for CGI programs */
char	cginame[256];		/* full name of webdir */
char	exename[256];		/* executable extension */
char	xfilename[256];
char	tfilename[256];
char	wfilename[256];
char	vfilename[256];
char	wordname[256];
char	idwordname[256];
char	ipaname[256];
char	syscom[1024];
char	*nulstr="";
const char *href;

#define MAXATTR	100
const char *sattr[MAXATTR];
int sattrcnt;

/* store string in sorted list */
int storeattr(const char *attr)
{
	int	i;

	if (sattrcnt==0)
		sattr[sattrcnt++] = attr;
	else {
		for (i=sattrcnt;(i>0) && (strcmp(attr,sattr[i-1])<0);i--)
			sattr[i] = sattr[i-1];
		sattr[i] = attr;
		sattrcnt++;
	}
	return(sattrcnt);
}

/* count levels below this node */
int countlevels(NSL_Item *item)
{
	NSL_Data	*data;
	int 		count=0,maxlevel=0;

	for (data=item->data;data!=NULL;data = data->next) {
		if (data->type==NSL_item_data)
			count = countlevels(data->first) + 1;
		else
			count = 1;
		if (count > maxlevel) maxlevel=count;
	}
			
	return(maxlevel);
}
	
/* count leaves below this node */
int countleaves(NSL_Item *item)
{
	NSL_Data	*data;
	int 		count=0;

	for (data=item->data;data!=NULL;data = data->next)
		if (data->type==NSL_item_data)
			count += countleaves(data->first);
		else
			count++;

	return(count);
}

/* dump an entry */
void entryprint(FILE *op,NSL_Item *item,int level)
{
	NSL_Data	*data;
	NSL_Attr	*attr;
	int		count;
	int		i;

	/* get lowest level count */
	count = countleaves(item);

	/* report this label */
	fprintf(op,"<td rowspan=%d valign=top><b>%s</b><br><sub><font color=\"#008000\">",count,item->label);
	sattrcnt=0;
	for (attr=item->attr;attr!=NULL;attr = attr->next) storeattr(attr->name);
	for (i=0;i<sattrcnt;i++)
		fprintf(op,"%s=%s ",sattr[i],GetAttrStringVal(item,sattr[i]));
	fprintf(op,"</sub></font></td>\n");

	/* report children */
	for (data=item->data;data!=NULL;data = data->next) {
		if (data->type==NSL_item_data) {
			if (data!=item->data) {
				/* 2nd and subsequent items on new rows */
				fprintf(op,"</tr>\n<tr>");
			}
			entryprint(op,data->first,level+1);
		}
		else {
			if (data==item->data) fprintf(op,"<td valign=top>");
			fprintf(op,"<font color=\"#0000C0\">%s</font> ",(char *)data->first);
			if (data->next==NULL) fprintf(op,"</td>");
		}
	}
}

/* process XML format file */
int	xmlprocess(char *xmlname,FILE *op)
{
	NSL_File 	inf;
	NSL_Doctype 	dct=NULL;
	NSL_Item 	*item;
	NSL_Bit		*bit;
	NSL_Data	*data;

	const char *	id;

	char		basename[256];
	char		afilename[256];
	char		tname[256];
	char		fullname[256];
	char		*p;
	int		nlevels;
	int		i;
	
	tname[0]='\0';

	if (*xmlname=='/')
		strcpy(fullname,xmlname);
	else
		sprintf(fullname,"/%s%s",basedir,xmlname);

	/* Initialise the NSL API */
	NSLInit(0);

	/* open file */
	if ((inf = SFopen(fullname,dct,NSL_read))==NULL)
		cgierror("could not find file '%s'",fullname);

	/* get the document type */
	dct = DoctypeFromFile(inf);

	id     = AttrUniqueName(dct,"ID",0);

	/* read first tag - should be <LEXICON> */
	if ((bit=GetNextBit(inf))==NULL)
		cgierror("Could not read <LEXICON> start tag");
	if ((bit->type!=NSL_start_bit) || (strcmp(bit->label,"LEXICON")))
		cgierror("Could not read <LEXICON> start tag");
	FreeBit(bit);

	fprintf(op,"<H2>%s:<BR>LEXICON Description</H2><P><HR>\n",fullname);

	/* Loop round reading components of the input text */
	while ((bit=GetNextBit(inf))!=NULL) {
		if (bit->type==NSL_start_bit) {
			item=ItemParse(inf,bit->value.item);

			/* should be an entry */
			if (strcmp(item->label,"ENTRY")==0) {

				fprintf(op,"<h3>%s</h3>\n",GetAttrStringVal(item,id));
				fprintf(op,"<p><table width=100%% border=1>\n");
				nlevels = countlevels(item);
				for (i=0;i<nlevels;i++) fprintf(op,"<th>%d</th>",i+1);
				fprintf(op,"\n<tr>");

				/* report children */
				for (data=item->data;data!=NULL;data = data->next) {
					if (data->type==NSL_item_data) {
						if (data!=item->data) {
							/* 2nd and subsequent items on new rows */
							fprintf(op,"</tr>\n<tr>");
						}
						entryprint(op,data->first,0);
					}
				}

				fprintf(op,"</tr>\n");
				fprintf(op,"</table>\n");
			}
			else
				cgierror("expected ENTRY, got %s\n",item->label);

			FreeItem(item);
		}
		FreeBit(bit);
	}

	
	SFclose(inf);
	return 0;
}

/* process template */
int	templateprocess(char *filename,FILE *op,char *xmlname)
{
	FILE	*ip;
	int	c,d;

	if ((ip=fopen(filename,"r"))==NULL)
		return -1;

	while ((c=getc(ip))!=EOF) {
		if (c=='{') {
			d = getc(ip);
			if (d=='}')
				xmlprocess(xmlname,op);
			else {
				putc(c,op);
				ungetc(d,ip);
			}
		}
		else
			putc(c,op);
	}
	fclose(ip);
	return 0;
}
				
/* main program */
void main(int argc, char *argv[])
{
	int	cl;
	char	*nval;
	char	*name,*value;
	FILE	*ip;
	FILE	*op;
	
	/* initialise directories */
#ifdef DOS
	strcpy(basedir,"c:/prosynth");
	strcpy(cgibase,"/cgi-bin/");
	strcpy(cginame,"/cgi-bin/webplay.exe");
	strcpy(exename,".exe");
#else
	strcpy(basedir,"/hdata/prosynth");
	strcpy(cgibase,"/htbin/");
	strcpy(cginame,"/htbin/webplay");
	strcpy(exename,"");
#endif

	/* open CGI stream */
	ip = cgiopen(&cl);

	strcpy(tfilename,"");
	strcpy(xfilename,"");

	/* loop through name value pairs */
	while (getnamevalue(ip,&cl,&nval)) {
		URLdecode(nval);
		name = strtok(nval,"=");
		if (!name || !*name)
			cgierror("Found NULL form field name");
		value = strtok(NULL,"");
		if (!value) value = nulstr;

		if (strcmp(name,"file")==0) {
			/* XML filename */
			strncpy(xfilename,value,256);
		}
		else if (strcmp(name,"template")==0) {
			/* template file name */
			strncpy(tfilename,value,256);
		}
		else if (strcmp(name,"word")==0) {
			/* name of a word */
			strncpy(wordname,value,256);
			strcpy(idwordname,wordname);
			strupr(idwordname);
		}
		else if (strcmp(name,"ipa")==0) {
			/* name of a word */
			strncpy(ipaname,value,256);
		}
		else if (*name) {
			cgierror("Unknown form field <code>%s = %s</code>",CHKSTR(name),CHKSTR(value));
		}
	}

	/* decide upon action */
	if (*wordname && *ipaname) {
		/* single word to process */
		chdir("/tmp");
		sprintf(wfilename,"lwXXXXXX");
		mktemp(wfilename);
		sprintf(vfilename,"lxXXXXXX");
		mktemp(vfilename);
		op = fopen(wfilename,"w");
		fprintf(op,"<?xml version=\"1.0\"?>\n<LEXICON>\n<ENTRY ID='%s'>\n",idwordname);
		fprintf(op,"<HW>%s</HW>\n",wordname);
		fprintf(op,"<PRONSEQ><PRON><IPA>%s</IPA></PRON></PRONSEQ>\n",ipaname);
		fprintf(op,"</ENTRY></LEXICON>\n");
		fclose(op);
		sprintf(syscom,"/src/xml/ucl/dev/lexparse %s %s",wfilename,vfilename);
		system(syscom);
		sprintf(xfilename,"//tmp/%s",vfilename);
	}
	if (*xfilename) {
		if (*tfilename)
			templateprocess(tfilename,stdout,xfilename);
		else {
			printf("<html><head><title>Contents of %s</title></head>\n",xfilename);
			printf("<body>\n");
			xmlprocess(xfilename,stdout);
			printf("</body></html>\n");
		}
	}	

	/* that's all folks */
	cgiclose(ip);
	exit(0);
}

