/* lookup - look up words in pronunciation dictionary */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#include "prondict.h"

static unsigned char left[256],right[256],stack[256];
static unsigned char *bfstart,*bfend,*bfpos;
static int bfsize,bfstack;

static void bfinit(unsigned char *buf,int size)
{
	unsigned char	*tab;
	int	c,count,i;

	/* read pair table */
	for (i=0;i<256;i++)
		left[i] = i;
	tab = buf;
	for (c=0;;) {
		count = *tab++;
		if (count > 127) {
			c += count - 127;
			count=0;
		}
		if (c==256) break;

		for (i=0;i<=count;i++,c++) {
			left[c] = *tab++;
			if (c!=(int)(left[c]))
				right[c] = *tab++;
		}
		if (c==256) break;
	}
	bfstart = tab;
	bfsize = size - (tab-buf);
	bfend = bfstart+bfsize;
	bfpos = bfstart;
	bfstack = 0;
}

static int bfnext()
{
	int	c;
	
	/* decompress */
	while (bfpos < bfend) {
		if (bfstack)
			c = stack[--bfstack];
		else {
			c = *bfpos++;
		}

		if (c==(int)(left[c]))
			return(c);
		else {
			stack[bfstack++] = right[c];
			stack[bfstack++] = left[c];
		}
	}

	return(-1);
}

static char *strsave(char *str)
{
	char	*ptr=malloc(strlen(str)+1);
	strcpy(ptr,str);
	return(ptr);
}

static int wordmatch(char *w1,char *w2)
{
	while (*w1 && (*w1!='#')) {
		if (!*w2) return(0);
		if (*w1!=*w2) {
			if (tolower(*w1)!=*w2)
				return(0);
		}
		w1++;
		w2++;
	}
	if (*w2) return(0);
	return(1);
}

int	lookup(char **words,int num,char **prons,int flags)
{
	int	c,i,j;
	char	lword[1024];
	char	lphon[1024];
	char	lbuf[1024];
	char	*p,*lp;
	int	n;
	int	more=num;

	for (i=0;i<num;i++) prons[i]=NULL;

	strcpy(lword,"");
	strcpy(lphon,"");

	bfinit(prondict,PRONDICTLEN);
	c=bfnext();
	while (more && (c!=ENDOFDICT)) {
		n = c;
		i = n;
		while (((c=bfnext())!=ENDOFDICT)&&(c>32)) lword[i++] = c;
		lword[i]='\0';

		n = c;
		i = n;
		while (((c=bfnext())!=ENDOFDICT)&&(c>32)) lphon[i++] = c;
		lphon[i]='\0';

		for (i=0;i<num;i++) {
			if (wordmatch(words[i],lword)) {
				if (prons[i]) {
					sprintf(lbuf,"%s\t%s",prons[i],lphon);
					free(prons[i]);
					prons[i]=strsave(lbuf);
				}
				else {
					prons[i]=strsave(lphon);
/*					more--; */
				}
			}
		}
	}

	for (i=0;i<num;i++) if (prons[i]==NULL) {
		strcpy(lphon,"");
		if (flags & 1) fprintf(stderr,"unknown word: '%s'\n",words[i]);
		for (j=0;words[i][j]!='\0';j++) switch(toupper(words[i][j])) {
		case 'A': strcat(lphon,"'eI"); break;
		case 'B': strcat(lphon,"'bi"); break;
		case 'C': strcat(lphon,"'si"); break;
		case 'D': strcat(lphon,"'di"); break;
		case 'E': strcat(lphon,"'i"); break;
		case 'F': strcat(lphon,"'ef"); break;
		case 'G': strcat(lphon,"'dZi"); break;
		case 'H': strcat(lphon,"'eItS"); break;
		case 'I': strcat(lphon,"'aI"); break;
		case 'J': strcat(lphon,"'dZeI"); break;
		case 'K': strcat(lphon,"'keI"); break;
		case 'L': strcat(lphon,"'el"); break;
		case 'M': strcat(lphon,"'em"); break;
		case 'N': strcat(lphon,"'en"); break;
		case 'O': strcat(lphon,"'@U"); break;
		case 'P': strcat(lphon,"'pi"); break;
		case 'Q': strcat(lphon,"'kju"); break;
		case 'R': strcat(lphon,"'A"); break;
		case 'S': strcat(lphon,"'es"); break;
		case 'T': strcat(lphon,"'ti"); break;
		case 'U': strcat(lphon,"'ju"); break;
		case 'V': strcat(lphon,"'vi"); break;
		case 'W': strcat(lphon,"'dVb@lju"); break;
		case 'X': strcat(lphon,"'eks"); break;
		case 'Y': strcat(lphon,"'waI"); break;
		case 'Z': strcat(lphon,"'zed"); break;
		case '0': strcat(lphon,"'zI@r@U"); break;
		case '1': strcat(lphon,"'wVn"); break;
		case '2': strcat(lphon,"'tu"); break;
		case '3': strcat(lphon,"'Tri"); break;
		case '4': strcat(lphon,"'fO"); break;
		case '5': strcat(lphon,"'faIv"); break;
		case '6': strcat(lphon,"'sIks"); break;
		case '7': strcat(lphon,"'sev@n"); break;
		case '8': strcat(lphon,"'eIt"); break;
		case '9': strcat(lphon,"'naIn"); break;
		}
		prons[i] = strsave(lphon);
	}

	for (i=0;i<num;i++) {
		if (p=strrchr(words[i],'#')) {
			n = atoi(p+1);
			if ((1<=n)&&(n<=9)) {
				p = strtok(prons[i],"\t");
				n--;
				while (p && *p && (n>0)) {
					lp = p;
					p = strtok(NULL,"\t");
					n--;
				}
				if (!p || !*p) p=lp;
				strcpy(lbuf,p);
				free(prons[i]);
				prons[i]=strsave(lbuf);
			}
		}
	}
	
	return(num);
}
	
#ifdef EMO
#define BLOCK	32
void main()
{
	int	i,j;
	char	buf[256];
	char	*words[BLOCK];
	char	*prons[BLOCK];

	i=0;
	while (fgets(buf,256,stdin)) {
		words[i] = strsave(strtok(buf,"\n"));
		i++;
		if (i==BLOCK) {
			lookup(words,i,prons,1);
			for (j=0;j<i;j++) {
				printf("%s\t%s\n",words[j],prons[j]);
				free(words[j]);
				free(prons[j]);
			}
			i=0;
		}
	}
	if (i>0) {
		lookup(words,i,prons,1);
		for (j=0;j<i;j++) {
			printf("%s\t%s\n",words[j],prons[j]);
			free(words[j]);
			free(prons[j]);
		}
		i=0;
	}
	exit(0);
}
#endif
