/* pretty -- traverse the content of an XML file with printpretty */

/* Mark Huckvale - University College London */

/* version 1.0 - March 1998 */

#define PROGNAME "pretty"
#define PROGVERS "1.0"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"

char		ofilename[256];
char		tfilename[256];

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	FILE		*op;
	NSL_File 	inf, outf;
	NSL_Doctype 	dct=NULL;
	NSL_Bit 	*bit;
	int		len;
	int		i;
	int		indent=0;
	int		compact=0;
	int		nl=0;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Ico:"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: pretty print XML file Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'c':
		compact=1;
		break;
	case 'o':
		strcpy(ofilename,optarg);
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-c) (-o output.xml) (file.xml)",PROGNAME);

	/* open file */
	if (optind < argc)
		inf = SFopen(argv[optind], dct, NSL_read);
	else
		inf = SFFopen(stdin, dct, NSL_read, "stdin");

	/* get the document type */
	dct = DoctypeFromFile(inf);

	/* Open the output NSGML stream using the same doctype declaration */
	if (ofilename[0]) {
		strcpy(tfilename,ofilename);
		strcat(tfilename,"_XXXXXX");
		mktemp(tfilename);
		op = fopen(tfilename,"w");
//		fprintf(op,"<?XML version='1.0' rmd='all'?>\n");
		if ((outf = SFFopen(op,dct,NSL_write_minimal,ofilename))==NULL)
			error("could not open output file '%s'",ofilename);
	}
	else {
		if ((outf = SFFopen(stdout, dct, NSL_write|NSL_write_plain, "stdout"))==NULL)
			error("could not open output stream");
	}

	/* Loop round reading components of the NSGML input text */
	nl=0;
	while ((bit=GetNextBit(inf))) {
		switch (bit->type) {

		case NSL_start_bit:
			/* print indent */
			if (nl) {
				PrintText(outf,"\n");
				for (i=0;i<indent;i++) PrintText(outf,"  ");
			}
			/* print tag */
			PrintBit(outf,bit);
			indent++;
			nl=1;
			break;
		case NSL_empty_bit:
			/* print indent */
			if (nl) {
				PrintText(outf,"\n");
				for (i=0;i<indent;i++) PrintText(outf,"  ");
			}
			/* print tag */
			PrintBit(outf,bit);
			nl=1;
			break;
		case NSL_text_bit:
			/* print indent */
			if (!compact) {
				PrintText(outf,"\n");
				for (i=0;i<indent;i++) PrintText(outf,"  ");
			}
			/* strip trailing whitespace */
			len=strlen(bit->value.body);
			while (strchr((char*)" \t\n",bit->value.body[len-1])) {
				bit->value.body[--len]='\000';
			}
			/* put out the word */
			PrintText(outf,bit->value.body);
			if (compact) nl=0; else nl=1;
			break;
		case NSL_end_bit:
			indent--;
			/* print indent */
			if (nl) {
				PrintText(outf,"\n");
				for (i=0;i<indent;i++) PrintText(outf,"  ");
			}
			/* print it no matter what */
			PrintBit(outf,bit);
			nl=1;
			break;
		default:
			PrintBit(outf,bit);
		}
		/* Bits are fixed but their contents are not freed
		       unless we do it ourselves */
		FreeBit(bit);
	}
	PrintText(outf,"\n");

	/* at the very end we need . . . */
	SFclose(inf);
	SFclose(outf);
	if (*ofilename) {
		remove(ofilename);
		rename(tfilename,ofilename);
	}
	return 0;
}
