/* builtin.c -- code for ProXML built-in functions */

#undef IAG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <math.h>
#include <time.h>
#include "lt-memory.h"
#include "nsl.h"
#include "nsllib.h"
#include "prx.h"
#include "builtin.h"
#include "process.h"
#include "url.h"

#ifdef SVR4
#define _stricmp strcasecmp
#endif

/*===============================================================*/

/* built-in functions */

void BI_tochar()
{
	Value val=pop();
	push(tochar(val));
}
void BI_tointeger()
{
	Value val=pop();
	push(tointeger(val));
}
void BI_tofloat()
{
	Value val=pop();
	push(tofloat(val));
}
void BI_tostring()
{
	Value val=pop();
	push(tostring(val));
}
void BI_numchild()
{
	NSL_Data *data;
	int	cnt=0;
	Value val;
	Value node = pop();

	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		for (data = ((NSL_Item *)node.u.v)->data;data!=NULL;data=data->next) {
			if ((data->type==NSL_item_data) || (data->type==NSL_text_data)) cnt++;
		}
	}
	val.type = INTEGER;
	val.u.i = cnt;
	push(val);
}
void BI_iselement()
{
	Value val;
	Value node = pop();

	val.type = INTEGER;
	if (node.type==NODE)
		val.u.i = 1;
	else
		val.u.i = 0;
	push(val);
}
void BI_istext()
{
	Value val;
	Value node = pop();

	val.type = INTEGER;
	if (node.type==STRING)
		val.u.i = 1;
	else
		val.u.i = 0;
	push(val);
}
void BI_output()
{
	Value val=tostring(pop());
	fputs(val.u.s,stdout);
	push(val);
}
void BI_element()
{
	NSL_Item *item;
	Value node=pop();
	Value val;
	val.type=STRING;
	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		item = (NSL_Item *)node.u.v;
		val.u.s = strsave(item->label);
	}
	else {
		val.u.s = strsave("");
	}
	push(val);
}
void BI_text()
{
	Value node=pop();
	Value val;
	val.type=STRING;
	if (node.type==STRING)
		val.u.s = strsave(node.u.s);
	else
		val.u.s = strsave("");
	push(val);
}
void BI_newattribute()
{
	NSL_ElementSummary defn;
	NSL_Item *item;
	Value node;
	Value val;
	char	*attrname;

	/* get attribute name */
	val = tostring(pop());

	/* get node */
	node = pop();

	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		item = (NSL_Item *)node.u.v;
		defn = item->defn;
		attrname = val.u.s;
		FindAttrSumAndName(gdct,&defn,item->label,&attrname,0);
/*		DeclareAttr(gdct,attrname,0,NSL_Dec_cdata,"",0,NSL_defval_implied,"",&defn,item->label); */
	}

	/* return something */
	push(val);
}
void BI_pow10()
{
	Value val=tofloat(pop());
	val.u.d = pow(10.0,val.u.d);
	push(val);
}
void BI_log10()
{
	Value val=tofloat(pop());
	if (val.u.d <= 1.0E-10)
		val.u.d = -10;
	else
		val.u.d = log10(val.u.d);
	push(val);
}

void BI_random()
{
        static int seed=0;
	int tmp_seed;
	Value	val;

	if (seed==0) seed=(int)time((time_t *)0);

	tmp_seed = 16807 * ( seed % 127773 ) - 2836 * (seed / 127773 );

        if (tmp_seed >= 0)
                seed = tmp_seed;
        else
                seed = tmp_seed + 2147483647;

	val.type = FLOAT;
	val.u.d = seed/2147483648.0;
	push(val);
}

void BI_exit()
{
	Value msg;
	Value code;
	msg = tostring(pop());
	code = tointeger(pop());
	if (*msg.u.s!='\0') fprintf(stderr,"%s",msg.u.s);
	exit(code.u.i);
}
void BI_format()
{
	Value	fmt;
	Value 	var;
	Value	out;
	char	*p;
	char	buff[256];

	buff[0]='\0';

	var = pop();
	fmt = tostring(pop());
	p = fmt.u.s;
	while (*p) {
		if (*p=='%') {
			p++;
			if (*p=='%') continue;
			while (*p && ((*p=='-')||isdigit(*p))) p++;
			if ((*p=='d')||(*p=='x')||(*p=='X')) {
				var = tointeger(var);
				sprintf(buff,fmt.u.s,var.u.i);
			}
			else if (*p=='c') {
				var = tochar(var);
				sprintf(buff,fmt.u.s,var.u.c);
			}
			else if ((*p=='f')||(*p=='g')||(*p=='e')) {
				var = tofloat(var);
				sprintf(buff,fmt.u.s,var.u.d);
			}
			else if (*p=='s') {
				var = tostring(var);
				sprintf(buff,fmt.u.s,var.u.s);
			}
			break;
		}
		p++;
	}
	out.u.s = strsave(buff);
	out.type = STRING;
	push(out);	
}

static NSL_Item * MyNewNullNSLItem(NSL_Doctype dct,char *name)
{
	char	elbuff[1024];

	if (dct->seenDTD) {
		/* assume want checking */
		return NewNullNSLItem(dct,name,0);
	}
	else {
		/* assume want freedom */
		sprintf(elbuff,"<%s></%s>",name,name);
		return GetItemFromString(elbuff,dct);
	}
}

void BI_appendchild()
{
	NSL_Item *item;
	NSL_Item *newitem;
	Value	node;
	Value	val;

	/* get element name */
	val = tostring(pop());

	/* get node */
	node = pop();

	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		item = (NSL_Item *)node.u.v;
		newitem = MyNewNullNSLItem(gdct,val.u.s);
		newitem->type = NSL_empty;
		AddItemToEnd(item,newitem);
		node.type=NODE;
		node.u.v=newitem;
	}
	else {
		node.type=NODE;
		node.u.v=NULL;
	}
	push(node);
}

void BI_newitem()
{
	NSL_Item *item;
	Value	val;
	Value	node;

	/* get element name */
	val = tostring(pop());

	/* create item */
	item = MyNewNullNSLItem(gdct,val.u.s);
	item->type = NSL_empty;

	/* return node pointer */
	node.type=NODE;
	node.u.v=item;

	push(node);
}

void BI_appenditem()
{
	NSL_Item *item;
	NSL_Item *newitem;
	Value	node;
	Value	child;

	/* get new item */
	child = pop();

	/* get parent node */
	node = pop();

	if ((node.type==NODE)&&(node.u.v!=NULL)&&(child.type==NODE)&&(child.u.v!=NULL)) {
		item = (NSL_Item *)node.u.v;
		newitem = CopyItem((NSL_Item *)child.u.v);
		AddItemToEnd(item,newitem);
		node.type = NODE;
		node.u.v = newitem;
	}
	else {
		node.type=NODE;
		node.u.v=NULL;
	}
	push(node);
}

void BI_deleteitem()
{
	Value	node;
	NSL_Item *item;
	NSL_Item *parent=NULL;
	NSL_Data *data,*last=NULL;

	/* get node to delete */
	node = pop();

	/* check node is OK */
	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		item = (NSL_Item *)node.u.v;
		/* get data holding this item */
		data = item->in;
		if (data!=NULL) {
			/* get parent holding this data */
			parent = data->in;
			for (data=parent->data;data!=NULL;data=data->next) {
				if (data->first==item) {
					/* found us */
					if (!last)
						parent->data = data->next;
					else
						last->next = data->next;
					FreeItem(item);
				}
				last=data;
			}
		}
	}

	/* return name of parent */
	node.type=NODE;
	node.u.v=parent;

	push(node);
}

static int testdata2(NSL_Data *data1,NSL_Data *data2,int faddr)
{
	Value	node;
	Value	flag;

	/* check for NULL */
	if (!data1||!data2) return(0);

	/* clear out return value */
	gretval.type=INTEGER;
	gretval.u.i = 0;

	/* push node1 */
	if (data1->type==NSL_item_data) {
		node.type = NODE;
		node.u.v = data1->first;
	}
	else {
		node.type = STRING;
		node.u.v = data1->first;
	}
	push(node);

	/* push node2 */
	if (data2->type==NSL_item_data) {
		node.type = NODE;
		node.u.v = data2->first;
	}
	else {
		node.type = STRING;
		node.u.v = data2->first;
	}
	push(node);

	/* call user function */
	perform(NULL,gdct,faddr,FUNCEND);

	/* remove nodes from stack */
	pop();
	pop();

	/* check return value */
	flag = tointeger(gretval);

	return(flag.u.i);
}

void BI_sortchildren()
{
	NSL_Item *item;
	NSL_Data *data,*ctmp,**ctab;
	Value	node;
	Value	faddr;
	int	i,j,nchild;

	/* get function address */
	faddr = tointeger(pop());

	/* get node */
	node = pop();

	/* get pointer to XML node */
	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		/* get item */
		item = (NSL_Item *)node.u.v;
		/* count children */
		nchild = 0;
		for (data=item->data;data!=NULL;data=data->next)
			if ((data->type==NSL_item_data) || (data->type==NSL_text_data)) nchild++;
		if (nchild==0) goto done;
		/* allocate a table */
		ctab = (NSL_Data **)calloc(nchild,sizeof(NSL_Data *));
		/* copy over data ptrs */
		nchild = 0;
		for (data=item->data;data!=NULL;data=data->next)
			if ((data->type==NSL_item_data) || (data->type==NSL_text_data))
				ctab[nchild++]=data;
		/* insertion sort */
		for (i=1;i<nchild;i++) {
			j=i;
			ctmp=ctab[j];
			while ((j>0)&&(testdata2(ctmp,ctab[j-1],faddr.u.i)<0)) {
				ctab[j]=ctab[j-1];
				j--;
			}
			ctab[j] = ctmp;
		}
		/* copy ptrs back */
		item->data = ctab[0];
		data = ctab[0];
		for (i=1;i<nchild;i++) {
			data->next=ctab[i];
			data=ctab[i];
		}
		data->next=NULL;
		/* free table */
		free(ctab);
	}
	else {
		node.type=NODE;
		node.u.v=NULL;
	}
done:
	push(node);
}

static int testnode(NSL_Item *item,int faddr)
{
	Value	node;
	Value	flag;

	/* check for NULL */
	if (!item) return(0);

	/* clear out return value */
	gretval.type=INTEGER;
	gretval.u.i = 0;

	/* push node */
	node.type = NODE;
	node.u.v = item;
	push(node);

	/* call user function */
	perform(NULL,gdct,faddr,FUNCEND);

	/* remove node from stack */
	pop();

	/* check return value */
	flag = tointeger(gretval);

	return(flag.u.i);
}

static NSL_Item * ancestor_search(NSL_Item *item,int faddr)
{
	NSL_Data *data;

	/* check for NULL */
	if (!item) return(NULL);

	/* try parent */
	data = item->in;
	if (!data || !data->in)
		return NULL;
	item = data->in;

	/* check for a match */
	if (testnode(item,faddr)) return(item);

	/* search up some more */
	return ancestor_search(item,faddr);
}

void BI_ancestor()
{
	NSL_Item *item;
	Value	node;
	Value	faddr;

	/* get function address */
	faddr = tointeger(pop());

	/* get node */
	node = pop();

	/* get pointer to XML node */
	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		/* test node parent */
		item = (NSL_Item *)node.u.v;
		node.u.v = ancestor_search(item,faddr.u.i);
	}
	else {
		node.type=NODE;
		node.u.v=NULL;
	}
	push(node);
}

/* search descendants of this node */
static NSL_Item * descendant_search(NSL_Item *item,int faddr)
{
	NSL_Data *data;

	/* check for NULL */
	if (!item) return(NULL);

	/* try each descendant */
	for (data = item->data;data!=NULL;data=data->next) {
		if (data->type==NSL_item_data) {
			item = data->first;
			/* check for a match */
			if (testnode(item,faddr)) return(item);
			/* search descendants */
			item = descendant_search(item,faddr);
			if (item) return (item);
		}
	}
	return NULL;
}

void BI_descendant()
{
	NSL_Item *item;
	Value	node;
	Value	faddr;

	/* get function address */
	faddr = tointeger(pop());

	/* get node */
	node = pop();

	/* get pointer to XML node */
	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		/* get item */
		item = (NSL_Item *)node.u.v;
		node.u.v = descendant_search(item,faddr.u.i);
	}
	else {
		node.type=NODE;
		node.u.v=NULL;
	}
	push(node);
}

static NSL_Data * prevdata(NSL_Data *data)
{
	NSL_Item *item = data->in;
	NSL_Data *pdata;

	if (item==NULL) return(NULL);
	for (pdata=item->data;(pdata!=NULL)&&(pdata->next!=data);pdata=pdata->next)
		/* skip */;
	return(pdata);
}

/* search descendants of this node - backwards! */
static NSL_Item * back_descendant_search(NSL_Item *item,int faddr)
{
	NSL_Data *data;
	NSL_Item *nitem;

	/* check for NULL */
	if (!item) return(NULL);

	/* find last descendant */
	for (data = item->data;(data!=NULL)&&(data->next!=NULL);data=data->next)
		/* skip */;

	/* test each child */
	for (;data!=NULL;data=prevdata(data)) {
		if (data->type==NSL_item_data) {
			nitem = data->first;
			/* search descendants */
			nitem = back_descendant_search(nitem,faddr);
			if (nitem) return (nitem);
			/* check for a match */
			nitem = data->first;
			if (testnode(nitem,faddr)) return(nitem);
		}
	}

	return NULL;
}

static NSL_Item * previous_search(NSL_Item *item,int faddr)
{
	NSL_Data *data;
	NSL_Item *nitem;

	/* check for NULL */
	if (!item) return(NULL);

	/* find the previous node */
	data = item->in;
	if (data==NULL) return(NULL);
	data = prevdata(data);
	for (;data!=NULL;data=prevdata(data)) {
		if (data->type==NSL_item_data) {
			/* get item */
			nitem = data->first;
			/* try each descendant */
			nitem = back_descendant_search(nitem,faddr);
			if (nitem) return (nitem);
			/* test node itself */
			nitem = data->first;
			if (testnode(nitem,faddr)) return(nitem);
		}
	}

	/* finished at this level - so go up */
	data = item->in;
	if ((data==NULL)||(data->in==NULL)) return(NULL);
	item = data->in;
/*	if (testnode(item,faddr)) return(item); -- ancestors not considered previous */
	return previous_search(item,faddr);
}

void BI_preceding()
{
	NSL_Item *item;
	Value	node;
	Value	faddr;

	/* get function address */
	faddr = tointeger(pop());

	/* get node */
	node = pop();

	/* get pointer to XML node */
	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		item=(NSL_Item *)node.u.v;
		node.u.v=previous_search(item,faddr.u.i);
	}
	else {
		node.type=NODE;
		node.u.v=NULL;
	}

	push(node);
}

static NSL_Item * following_search(NSL_Item *item,int faddr)
{
	NSL_Data *data;
	NSL_Item *nitem;

	/* check for NULL */
	if (!item) return(NULL);

	/* find the following node */
	data = item->in;
	if (data==NULL) return(NULL);
	data = data->next;
	for (;data!=NULL;data=data->next) {
		if (data->type==NSL_item_data) {
			/* get item */
			nitem = data->first;
			if (testnode(nitem,faddr)) return(nitem);
			/* try each descendant */
			nitem = descendant_search(nitem,faddr);
			if (nitem) return (nitem);
		}
	}

	/* finished at this level - so go up */
	data = item->in;
	if (data==NULL) return(NULL);
	return following_search(data->in,faddr);
}

void BI_following()
{
	NSL_Item *item;
	Value	node;
	Value	faddr;

	/* get function address */
	faddr = tointeger(pop());

	/* get node */
	node = pop();

	/* get pointer to XML node */
	if ((node.type==NODE)&&(node.u.v!=NULL)) {
		item=(NSL_Item *)node.u.v;
#if 0
		/* try each descendant first */
		if ((node.u.v = descendant_search(item,faddr.u.i))==NULL)
#endif
			/* else try following */
			node.u.v=following_search(item,faddr.u.i);
	}
	else {
		node.type=NODE;
		node.u.v=NULL;
	}
	push(node);
}

/* text stream buffering */
struct FILE16BUFF {
	FILE16 		*fp;
	unsigned char 	buff[4096];
	int		pos,len;
	int		iswrite;
};

/* open a file */
static struct FILE16BUFF * FILE16BUFF_open(const char * name,const char * mode)
{
	struct FILE16BUFF *fb;

	fb=(struct FILE16BUFF *)malloc(sizeof(struct FILE16BUFF));
	fb->fp = url_open(name,NULL,mode,NULL);
	if (*mode=='w') {
		fb->iswrite=1;
		fb->pos = 0;
		fb->len = sizeof(fb->buff);
	}
	else {
		fb->iswrite=0;
		fb->pos = 0;
		fb->len = -1;
	}
	return(fb);
}
/* fill a block */
static void FILE16BUFF_fill(struct FILE16BUFF *fb)
{
	fb->len = Readu(fb->fp,fb->buff,sizeof(fb->buff));
	fb->pos = 0;
}
/* flush a block */
static int FILE16BUFF_flush(struct FILE16BUFF *fb)
{
	int cnt = Writeu(fb->fp,fb->buff,fb->pos);
	fb->pos = 0;
	return(cnt);
}
/* close a file */
static void FILE16BUFF_close(struct FILE16BUFF *fb)
{
	if (fb->iswrite) FILE16BUFF_flush(fb);
	Fclose(fb->fp);
	free(fb);
}
/* get a character */
static int FILE16BUFF_getc(struct FILE16BUFF *fb)
{
	if (fb->len==0)
		return(-1);
	else if (fb->pos < fb->len)
		return(fb->buff[fb->pos++]);
	else {
		FILE16BUFF_fill(fb);
		return(FILE16BUFF_getc(fb));
	}
}
/* put a character */
static int FILE16BUFF_putc(struct FILE16BUFF *fb,int c)
{
	if (fb->pos==fb->len) {
		if (FILE16BUFF_flush(fb)==0) return(0);
	}
	fb->buff[fb->pos++] = c;
	return(1);
}
/* put a string */
static int FILE16BUFF_puts(struct FILE16BUFF *fb,char *s)
{
	int	cnt=0;
	while (*s) {
		while (*s && (fb->pos < fb->len)) {
			fb->buff[fb->pos++] = *s++;
			cnt++;
		}
		if (fb->pos==fb->len) {
			if (FILE16BUFF_flush(fb)==0) return(0);
		}
	}
	return(cnt);
}

/* text stream IO */
void BI_streamopen()
{
	Value	fname;
	Value	fmode;
	Value	fid;
	struct FILE16BUFF	*fb;

	/* get open mode */
	fmode = tointeger(pop());

	/* get file name */
	fname = tostring(pop());

	fid.type = INTEGER;
	if (fmode.u.i) {
		/* open for writing */
		fb = FILE16BUFF_open(fname.u.s,"w");
	}
	else {
		/* open for reading */
		fb = FILE16BUFF_open(fname.u.s,"r");
	}
	if (fb->fp==NULL)
		fid.u.i = 0;
	else
		fid.u.i = (int)fb;

	push(fid);
}

void BI_streamclose()
{
	Value	fid;

	/* get file descriptor */
	fid = tointeger(pop());
	if (fid.u.i!=0) FILE16BUFF_close((struct FILE16BUFF *)fid.u.i);

	fid.u.i=0;
	push(fid);
}

void BI_streamget()
{
	Value	fid;
	Value	ch;
	
	/* get file descriptor */
	fid = tointeger(pop());

	ch.type = INTEGER;
	if (fid.u.i!=0)
		ch.u.i = FILE16BUFF_getc((struct FILE16BUFF *)fid.u.i);
	else
		ch.u.i = -1;

	push(ch);
}

void BI_streamgetline()
{
	Value	fid;
	Value	str;
	unsigned char buff[4096];
	int 	len;
	int	c;

	/* get file descriptor */
	fid = tointeger(pop());

	str.type=STRING;
	if (fid.u.i!=0) {
		len=0;
		while ((len < 4094)&&((c=FILE16BUFF_getc((struct FILE16BUFF *)fid.u.i))!=-1)&&(c!='\n')) {
			buff[len++] = c;
		}
		if ((len > 0)&&(buff[len-1]=='\r')) len--;
		if (c=='\n') buff[len++]='\n';
		buff[len]='\0';
		str.u.s = strsave(buff);
	}
	else {
		str.u.s = strsave("");
	}

	push(str);
}

/* put a character, number or string */
void BI_streamput()
{
	char	sbuf[256];
	Value	fid;
	Value	val;
	Value	cnt;

	/* initialise return count */
	cnt.type = INTEGER;
	cnt.u.i = 0;

	/* get value to output */
	val = pop();

	/* get file descriptor */
	fid = tointeger(pop());
	if (fid.u.i==0) {
		push(cnt);
		return;
	}

	switch (val.type) {
	case CHAR:	cnt.u.i = FILE16BUFF_putc((struct FILE16BUFF *)fid.u.i,val.u.c);
			break;
	case INTEGER:	sprintf(sbuf,"%d",val.u.i);
			cnt.u.i = FILE16BUFF_puts((struct FILE16BUFF *)fid.u.i,sbuf);
			break;
	case FLOAT:	sprintf(sbuf,"%g",val.u.d);
			cnt.u.i = FILE16BUFF_puts((struct FILE16BUFF *)fid.u.i,sbuf);
			break;
	case STRING:	if (val.u.s==NULL)
				cnt.u.i = FILE16BUFF_puts((struct FILE16BUFF *)fid.u.i,"(nil)");
			else
				cnt.u.i = FILE16BUFF_puts((struct FILE16BUFF *)fid.u.i,val.u.s);
			break;
	case NODE:	sprintf(sbuf,"ND%08X",val.u.i);
			cnt.u.i = FILE16BUFF_puts((struct FILE16BUFF *)fid.u.i,sbuf);
			break;
	}

	push(cnt);
}


/* parse a string into a node structure */
void BI_parse()
{
	Value	str;
	Value	typ;
	Value	elm;
	Value	node;
	NSL_Item *item;
	NSL_Data *data;
	char	*hold;
	char	*p;

	elm = tostring(pop());
	typ = tostring(pop());
	str = tostring(pop());

	node.type=NODE;
	node.u.v=NULL;

	if (strcmp(typ.u.s,"xml")==0) {
		if (elm.u.s[0]!='\0') {
			item = MyNewNullNSLItem(gdct,elm.u.s);
			item->type = NSL_empty;
			AddItemToEnd(item,GetItemFromString(str.u.s,gdct));
			node.u.v = (void *)item;
		}
		else
			node.u.v = (void *)GetItemFromString(str.u.s,gdct);
	}
	else if (strcmp(typ.u.s,"text")==0) {
		if (elm.u.s[0]!='\0')
			item = MyNewNullNSLItem(gdct,elm.u.s);
		else
			item = MyNewNullNSLItem(gdct,"TEXT");
		item->type = NSL_empty;
		hold = strsave(str.u.s);
		p = strtok(hold," \t\r\n");
		if (p && *p) {
#ifdef SGI
			data = (NSL_Data *)calloc(1,sizeof(NSL_Data));
			data->type = NSL_text_data;
			data->first = strsave(p);
#else
			data = NewTextNSLData(gdct,p,strlen(p)+1,TRUE,NULL,FALSE);
#endif
			InstallData(item,data);
			p = strtok(NULL," \t\r\n");
			while (p && *p) {
#ifdef SGI
				data->next = (NSL_Data *)calloc(1,sizeof(NSL_Data));
				data->next->type = NSL_text_data;
				data->next->first = strsave(p);
#else
				data->next = NewTextNSLData(gdct,p,strlen(p)+1,TRUE,NULL,FALSE);
#endif
				data->next->in = data->in;
				data = data->next;
				p = strtok(NULL," \t\r\n");
			}
		}
		node.u.v = (void *)item;
		free(hold);
	}

	push(node);
}
