/* process.c -- code for XML processing interpreter */

#undef IAG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <math.h>
#include "lt-memory.h"
#include "nsl.h"
#include "nsllib.h"
#include "prx.h"
#include "process.h"
#include "url.h"

#ifdef SVR4
#define _stricmp strcasecmp
#endif

/* global memory for functions */
#define MAXFUNCTIONS 1000
struct function_pc {
	int	pc;
	struct function_pc *next;
};
struct function_rec {
	char	*name;
	struct function_pc *first;
};
struct function_rec functab[MAXFUNCTIONS];
int funcpos;

/* operation stack */
#define MAXSTACK 1000
Value	stack[MAXSTACK];
int	sp;
int	globalpc;

/* allocate memory for code */
int codeexpand(int v)
{
	if (code==NULL) {
		code = calloc(CODECHUNKSIZE,sizeof(int));
		codesize = CODECHUNKSIZE;
	}
	else {
		code = realloc(code,2*codesize*sizeof(int));
		codesize *= 2;
	}
	if (code==NULL)
		prxerror("out of memory","");
	code[codepos++] = v;
	return(v);
}

/* add a function to function table */
int addfunction(char *name,int pc)
{
	int	i;
	struct function_pc *fpc;
	struct function_pc *ftmp;

	fpc = (struct function_pc *)malloc(sizeof(struct function_pc));
	if (fpc==NULL)
		prxerror("out of memory","");
	fpc->pc = pc;
	fpc->next = NULL;

	for (i=0;i<funcpos;i++) {
		if (strcmp(name,functab[i].name)==0) {
			if (functab[i].first==NULL)
				functab[i].first = fpc;
			else {
				ftmp = functab[i].first;
				while (ftmp->next!=NULL) ftmp=ftmp->next;
				ftmp->next = fpc;
			}
			return(0);
		}
	}
	if (funcpos < MAXFUNCTIONS) {
		functab[funcpos].name = strsave(name);
		functab[funcpos].first = fpc;
		funcpos++;
	}
	else
		prxerror("too many functions","");
	return(1);
}

int printop(int op1,int op2,int op3,int op4)
{
	switch (op1) {
	case ERROR:	printf("ERROR");	return(1);
	case END:	printf("END");		return(1);
	case FUNCSTART:	if (op2 >= 0)
				printf("FUNCSTART %s",symtab[op2]->name);
			else
				printf("FUNCSTART [%d]",op2);
			return(2);
	case FUNCEND:	printf("FUNCEND");	return(1);
	case VARSTORE:	printf("VARSTORE %d",op2);	return(2);
	case VARLOAD:	printf("VARLOAD %d",op2);	return(2);
	case VECINIT1:	printf("VECINIT1 %d",op2);	return(2);
	case GVECINIT1:	printf("GVECINIT1 %d",op2);	return(2);
	case VELLOAD1:	printf("VELLOAD1 %d",op2);	return(2);
	case GVELLOAD1:	printf("GVELLOAD1 %d",op2);	return(2);
	case VECINIT2:	printf("VECINIT2 %d",op2);	return(2);
	case GVECINIT2:	printf("GVECINIT2 %d",op2);	return(2);
	case VELLOAD2:	printf("VELLOAD2 %d",op2);	return(2);
	case GVELLOAD2:	printf("GVELLOAD2 %d",op2);	return(2);
	case GVARSTORE:	printf("GVARSTORE %d",op2);	return(2);
	case GVARLOAD:	printf("GVARLOAD %d",op2);	return(2);
	case NODELOADROOT: printf("NODELOADROOT");	return(1);
	case NODELOADTHIS: printf("NODELOADTHIS");	return(1);
	case NODELOADVAR: printf("NODELOADVAR %d",op2);	return(2);
	case GNODELOADVAR: printf("GNODELOADVAR %d",op2);	return(2);
	case NODELOADANY: printf("NODELOADANY");	return(1);
	case NODELOADNIL: printf("NODELOADNIL");	return(1);
	case NODELOAD:	printf("NODELOAD %s",symtab[op2]->name); return(2);
	case NODESTORE:	printf("NODESTORE %d",op2);	return(2);
	case GNODESTORE:	printf("GNODESTORE %d",op2);	return(2);
	case ATTRSTORE:	printf("ATTRSTORE %s",symtab[op2]->name); return(2);
	case ATTRLOAD:	printf("ATTRLOAD %s",symtab[op2]->name); return(2);
	case CONLOAD:	printf("CONLOAD %s",symtab[op2]->name); return(2);
	case PUSHLVAL:	printf("PUSHLVALVAR %d",op2);	return(2);
	case GPUSHLVAL:	printf("GPUSHLVALVAR %d",op2);	return(2);
	case LVALVARCOPY:	printf("LVALVARCOPY"); return(1);
	case LVALVECCOPY1:	printf("LVALVECCOPY1"); return(1);
	case LVALVECCOPY2:	printf("LVALVECCOPY2"); return(1);
	case LVALVARSTORE:	printf("LVALVARSTORE"); return(1);
	case LVALVECSTORE1:	printf("LVALVECSTORE1"); return(1);
	case LVALVECSTORE2:	printf("LVALVECSTORE2"); return(1);
	case LVALVECINIT:	printf("LVALVECINIT %d",op2); return(2);
	case LVALNODESTORE:	printf("LVALNODESTORE"); return(1);
	case LVALNODEVECSTORE1:	printf("LVALNODEVECSTORE1"); return(1);
	case LVALNODEVECSTORE2:	printf("LVALNODEVECSTORE2"); return(1);
	case ADD:	printf("ADD");			return(1);
	case SUBTRACT:	printf("SUBTRACT");		return(1);
	case MULTIPLY:	printf("MULTIPLY");		return(1);
	case DIVIDE:	printf("DIVIDE");		return(1);
	case NEGATE:	printf("NEGATE");		return(1);
	case CONCATENATE: printf("CONCATENATE");	return(1);
	case IF:	printf("IF");			return(1);
	case CONDEND:	printf("CONDEND");		return(1);
	case JZ:	printf("JZ %d",op2);		return(2);
	case JNZ:	printf("JNZ %d",op2);		return(2);
	case JMP:	printf("JMP %d",op2);		return(2);
	case TESTEQ:	printf("TESTEQ");		return(1);
	case TESTNE:	printf("TESTNE");		return(1);
	case TESTLT:	printf("TESTLT");		return(1);
	case TESTLE:	printf("TESTLE");		return(1);
	case TESTGT:	printf("TESTGT");		return(1);
	case TESTGE:	printf("TESTGE");		return(1);
	case LOGICALOR:	printf("LOGICALOR");		return(1);
	case LOGICALAND: printf("LOGICALAND");		return(1);
	case LOGICALNOT: printf("LOGICALNOT");		return(1);
	case GSTACKRESERVE: printf("GSTACKRESERVE %d",op2); return(2);
	case STACKRESERVE: printf("STACKRESERVE %d",op2); return(2);
	case STACKFREE:	printf("STACKFREE");		return(1);
	case ASCEND:	printf("ASCEND");		return(1);
	case DESCEND:	printf("DESCEND");		return(1);
	case DESCENDSEL: printf("DESCENDSEL");		return(1);
	case DESCENDEACH: printf("DESCENDEACH");	return(1);
	case DESCENDALL: printf("DESCENDALL");		return(1);
	case DESCENDREC: printf("DESCENDREC");		return(1);
	case FUNCTIONCALL: printf("FUNCTIONCALL %s",symtab[op2]->name); return(2);
	case RETURN:	printf("RETURN");		return(1);
	case NTESTEQ:	printf("NTESTEQ");		return(1);
	case NTESTNE:	printf("NTESTNE");		return(1);
	case POP:	printf("POP");			return(1);
	case FOREACH:	printf("FOREACH %d %d %d",op2,op3,op4); return(4);
	case GFOREACH:	printf("GFOREACH %d %d %d",op2,op3,op4); return(4);
	case PATTSAVE:	printf("PATTSAVE");		return(1);
	case PATTEND:	printf("PATTEND");		return(1);
	case DUPSTACK2:	printf("DUPSTACK2");		return(1);
	case SIZEOF:	printf("SIZEOF %d",op2);	return(2);
	case GSIZEOF:	printf("GSIZEOF %d",op2);	return(2);
	case ARRSIZE:	printf("ARRSIZE %d",op2);	return(2);
	case GARRSIZE:	printf("GARRSIZE %d",op2);	return(2);
	case ARRSIZE2:	printf("ARRSIZE2 %d",op2);	return(2);
	case GARRSIZE2:	printf("GARRSIZE2 %d",op2);	return(2);
	case LOADTRUE:	printf("LOADTRUE");		return(1);
	case LOADFALSE:	printf("LOADFALSE");		return(1);
	case ASSERT:	printf("ASSERT %d",op2);	return(2);
	case ASSERTVAL:	printf("ASSERTVAL %d",op2);	return(2);
	default:	printf("Unknown code=%d",op1);	return(1);
	}
}

/* dump code (for debugging) */
void dumpcode()
{
	int	i;

	printf("Code Table:\n");
	for (i=0;i<codepos;) {
		printf("%3d. (%2d) ",i,code[i]);
		i += printop(code[i],code[i+1],code[i+2],code[i+3]);
		printf("\n");
	}
}

/* push a value */
void push(Value val)
{
	if (sp < MAXSTACK)
		stack[sp++] = val;
	else
		prxerror("Stack overflow","");
}

Value pop()
{
	if (sp <= 0)
		prxerror("Stack underflow","");
	return stack[--sp];
}

Value tochar(Value val)
{
	switch (val.type) {
	case CHAR:	break;
	case INTEGER:	val.u.c = val.u.i & 0xFF; break;
	case FLOAT:	val.u.c = ((int)val.u.d) & 0xFF; break;
	case STRING:	val.u.c = *val.u.s; break;
	}
	val.type = CHAR;
	return(val);
}
Value tointeger(Value val)
{
	switch (val.type) {
	case CHAR:	val.u.i = val.u.c; break;
	case INTEGER:	break;
	case FLOAT:	val.u.i = (int)val.u.d; break;
	case STRING:	if ((val.u.s[0]=='-')||isdigit(val.u.s[0]))
				val.u.i = atoi(val.u.s);
			else
				fprintf(stderr,"Warning: integer conversion failed on '%s' at PC:%04d\n",val.u.s,globalpc);
			break;
	}
	val.type = INTEGER;
	return(val);
}
Value tofloat(Value val)
{
	switch (val.type) {
	case CHAR:	val.u.d = (double)val.u.c; break;
	case INTEGER:	val.u.d = (double)val.u.i; break;
	case FLOAT:	break;
	case STRING:	if ((val.u.s[0]=='-')||isdigit(val.u.s[0]))
				val.u.d = atof(val.u.s);
			else
				fprintf(stderr,"Warning: float conversion failed on '%s' at PC:%04d\n",val.u.s,globalpc);
			break;
	}
	val.type = FLOAT;
	return(val);
}
Value tostring(Value val)
{
	char	sbuf[32];

	switch (val.type) {
	case CHAR:	sbuf[0]=val.u.c; sbuf[1]='\0'; break;
	case INTEGER:	sprintf(sbuf,"%d",val.u.i); break;
	case FLOAT:	sprintf(sbuf,"%g",val.u.d); break;
	case STRING:	if (val.u.s==NULL) strcpy(sbuf,"(nil)"); else return(val);
	case NODE:	sprintf(sbuf,"ND%08X",val.u.i); break;
	default:	strcpy(sbuf,"(nil)"); break;
	}
	val.u.s = strsave(sbuf);
	val.type = STRING;
	return(val);
}
int isinteger(Value val)
{
	switch (val.type) {
	case CHAR:	return(1);
	case INTEGER:	return(1);
	case FLOAT:	return((val.u.d==(double)((int)val.u.d))?1:0);
	case STRING:	return(((double)atoi(val.u.s)==atof(val.u.s))?1:0);
	}
	return(0);
}
Value tonumber(Value val)
{
	switch (val.type) {
	case CHAR:	return(tointeger(val));
	case INTEGER:	return(val);
	case FLOAT:	return(val);
	case STRING:	return(isinteger(val)?tointeger(val):tofloat(val));
	}
	return(val);
}

/* compare two values */
int valcompare(Value val1,Value val2)
{
	if (val1.type==val2.type) {
		switch (val1.type) {
		case CHAR:	return((int)val1.u.c - (int)val2.u.c);
		case INTEGER:	return(val1.u.i - val2.u.i);
		case FLOAT:	if (val1.u.d < val2.u.d)
					return(-1);
				else if (val1.u.d > val2.u.d)
					return(1);
				else
					return(0);
		case STRING:	return(strcmp(val1.u.s,val2.u.s));
		}
	}
	else {
		val1 = tonumber(val1);
		val2 = tonumber(val2);
		if ((val1.type==INTEGER) && (val2.type==INTEGER)) {
			return(val1.u.i - val2.u.i);
		}
		else {
			val1 = tofloat(val1);
			val2 = tofloat(val2);
			if (val1.u.d < val2.u.d)
				return(-1);
			else if (val1.u.d > val2.u.d)
				return(1);
			else
				return(0);
		}
	}
	return(0); /* should be impossible */
}

/* current function offset into stack */
int fnoff=0;

/* global function return value */
Value gretval;

/* root address */
NSL_Item *groot;

/* perform a function on a single item */
int perform(NSL_Item *item,NSL_Doctype dct,int pc,int final)
{
	int	op;
	char	*str;
	Value	val,val2,node,node2;
	const char *attr;
	char	sbuf[4096];
	char	*s;
	const char *cs;
	int	addr;
	int	stoff;
	NSL_Data *data;
	NSL_Item *itemp;
/*	Value	(*func)(); */
	int	entry;
	int	i,pccode,pcexit;
	int	sp1,sp2;
	int	idx;

#ifdef IAG
	int	j;
printf("perform(%d,%d)\n",pc,final);
#endif	
	while ((op=code[pc++])!=final) {
#ifdef IAG
printf("Stack: ");
for (i=0;i<sp;i++) {
	if (stack[i].type==ARRAY) {
		printf("[");
		for (j=0;(j<5)&&(j<stack[i].u.a->ndim1*stack[i].u.a->ndim2);j++) {
			val = tostring(stack[i].u.a->ptr[j]);
			printf("'%s' ",val.u.s);
		}
		printf("] ");
	}	
	else {
		val=tostring(stack[i]);
		printf("'%s' ",val.u.s);
	}
}
printf("\nfnoff=%d pc=%d (%3d) ",fnoff,pc-1,op);
printop(op,code[pc],code[pc+1],code[pc+2]);
printf("\n");
#endif
	globalpc = pc;

	switch (op) {
	case ERROR:
		prxerror("Found ERROR code","");
		break;
	case END:
		return(0);
	case VARSTORE:
		stoff = fnoff + code[pc++];
		stack[stoff] = pop();
		break;
	case GVARSTORE:
		stoff = code[pc++];
		stack[stoff] = pop();
		break;
	case VARLOAD:
		stoff = fnoff + code[pc++];
		push(stack[stoff]);
		break;
	case GVARLOAD:
		stoff = code[pc++];
		push(stack[stoff]);
		break;
	case VECINIT1:
		stoff = fnoff + code[pc++];
		val = tointeger(pop());
		stack[stoff].type=ARRAY;
		stack[stoff].u.a = (struct array_rec *)calloc(1,sizeof(struct array_rec));
		stack[stoff].u.a->ndim1=1;
		stack[stoff].u.a->ndim2=val.u.i;
		stack[stoff].u.a->ptr=(struct Value_rec *)calloc(val.u.i,sizeof(struct Value_rec));
		break;
	case GVECINIT1:
		stoff = code[pc++];
		val = tointeger(pop());
		stack[stoff].type=ARRAY;
		stack[stoff].u.a = (struct array_rec *)calloc(1,sizeof(struct array_rec));
		stack[stoff].u.a->ndim1=1;
		stack[stoff].u.a->ndim2=val.u.i;
		stack[stoff].u.a->ptr=(struct Value_rec *)calloc(val.u.i,sizeof(struct Value_rec));
		break;
	case VECINIT2:
		stoff = fnoff + code[pc++];
		val = tointeger(pop());
		val2 = tointeger(pop());
		stack[stoff].type=ARRAY;
		stack[stoff].u.a = (struct array_rec *)calloc(1,sizeof(struct array_rec));
		stack[stoff].u.a->ndim1=val2.u.i;
		stack[stoff].u.a->ndim2=val.u.i;
		stack[stoff].u.a->ptr=(struct Value_rec *)calloc(val2.u.i*val.u.i,sizeof(struct Value_rec));
		break;
	case GVECINIT2:
		stoff = code[pc++];
		val = tointeger(pop());
		val2 = tointeger(pop());
		stack[stoff].type=ARRAY;
		stack[stoff].u.a = (struct array_rec *)calloc(1,sizeof(struct array_rec));
		stack[stoff].u.a->ndim1=val2.u.i;
		stack[stoff].u.a->ndim2=val.u.i;
		stack[stoff].u.a->ptr=(struct Value_rec *)calloc(val2.u.i*val.u.i,sizeof(struct Value_rec));
		break;
	case VELLOAD1:
		stoff = fnoff + code[pc++];
		val2 = tointeger(pop());
		idx = val2.u.i;
		if ((0 <= idx) && (idx < stack[stoff].u.a->ndim2))
			val = stack[stoff].u.a->ptr[idx];
		else {
			val.type = INTEGER;
			val.u.i=0;
		}
		push(val);
		break;
	case GVELLOAD1:
		stoff = code[pc++];
		val2 = tointeger(pop());
		idx = val2.u.i;
		if ((0 <= idx) && (idx < stack[stoff].u.a->ndim2))
			val = stack[stoff].u.a->ptr[idx];
		else {
			val.type = INTEGER;
			val.u.i=0;
		}
		push(val);
		break;
	case VELLOAD2:
		stoff = fnoff + code[pc++];
		val2 = tointeger(pop());
		idx = val2.u.i;
		val2 = tointeger(pop());
		idx += val2.u.i*stack[stoff].u.a->ndim2;
		if ((0 <= idx) && (idx < (stack[stoff].u.a->ndim1*stack[stoff].u.a->ndim2)))
			val = stack[stoff].u.a->ptr[idx];
		else {
			val.type = INTEGER;
			val.u.i=0;
		}
		push(val);
		break;
	case GVELLOAD2:
		stoff = code[pc++];
		val2 = tointeger(pop());
		idx = val2.u.i;
		val2 = tointeger(pop());
		idx += val2.u.i*stack[stoff].u.a->ndim2;
#ifdef IAG
printf("***GET idx=%d ndim1=%d ndim2=%d\n",idx,stack[stoff].u.a->ndim1,stack[stoff].u.a->ndim2);
#endif
		if ((0 <= idx) && (idx < (stack[stoff].u.a->ndim1*stack[stoff].u.a->ndim2)))
			val = stack[stoff].u.a->ptr[idx];
		else {
			val.type = INTEGER;
			val.u.i=0;
		}
		push(val);
		break;
	case NODELOADROOT:
		node.type = NODE;
		node.u.v = groot;
		push(node);
		break;
	case NODELOADTHIS:
		node.type = NODE;
		node.u.v = item;
		push(node);
		break;
	case NODELOADVAR:
		stoff = fnoff + code[pc++];
		push(stack[stoff]);
		break;
	case GNODELOADVAR:
		stoff = code[pc++];
		push(stack[stoff]);
		break;
	case NODELOADANY:
		val.type = STRING;
		val.u.s = strsave("");
		push(val);
		break;
	case NODELOADNIL:
		node.type = NODE;
		node.u.v = NULL;
		push(node);
		break;
	case NODELOAD:
		val.type = STRING;
		val.u.s = symtab[code[pc++]]->name;
		push(val);
		break;
	case NODESTORE:
		stoff = fnoff + code[pc++];
		stack[stoff] = pop();
		break;
	case GNODESTORE:
		stoff = code[pc++];
		stack[stoff] = pop();
		break;
	case ATTRSTORE:
		/* get expression */
		val = pop();
		switch (val.type) {
		case CHAR:	sprintf(sbuf,"%c",val.u.c); break;
		case INTEGER:	sprintf(sbuf,"%d",val.u.i); break;
		case FLOAT:	sprintf(sbuf,"%g",val.u.d); break;
		case STRING:	if (val.u.s!=NULL)
					strncpy(sbuf,val.u.s,sizeof(sbuf)-1);
				else
					strcpy(sbuf,"NIL");
				break;
		default:	strcpy(sbuf,"NIL"); break;
		}
		/* get attribute name */
		val = pop();
		val = tostring(val);
		str = val.u.s;
		attr = AttrUniqueName(dct,str,0);
		if (!attr)
			prxerror("could not find attribute named '%s'",str);
		/* get node */
		node = pop();
		if ((node.type==NODE)&&(node.u.v!=NULL))
			PutAttrVal((NSL_Item *)node.u.v,attr,strsave(sbuf));
		break;
	case ATTRLOAD:
		val = pop();
		val = tostring(val);
		str = val.u.s;
		attr = AttrUniqueName(dct,str,0);
		if (!attr)
			prxerror("could not find attribute named '%s'",str);
		node = pop();
		if ((node.type==NODE)&&(node.u.v!=NULL)) {
			cs = GetAttrVal((NSL_Item *)node.u.v,attr);
			if (cs!=NULL)
				val.u.s = strsave(cs);
			else
				val.u.s = strsave("");
		}
		else
			val.u.s = strsave("");
		val.type = STRING;
		push(val);
		break;
	case CONLOAD:
		val = symtab[code[pc++]]->val;
		push(val);
		break;
	case NAMELOAD:
		val.type=STRING;
		val.u.s = strsave(symtab[code[pc++]]->name);
		push(val);
		break;
	case LOADTRUE:
		val.type=INTEGER;
		val.u.i=1;
		push(val);
		break;
	case LOADFALSE:
		val.type=INTEGER;
		val.u.i=0;
		push(val);
		break;
	case PUSHLVAL:
		val.type=INTEGER;
		val.u.i=fnoff+code[pc++];
		push(val);
		break;
	case GPUSHLVAL:
		val.type=INTEGER;
		val.u.i=code[pc++];
		push(val);
		break;
	case LVALVARCOPY:
		val = stack[sp-1];
		stoff = val.u.i;
		push(stack[stoff]);
		break;
	case LVALVECCOPY1:
		val = stack[sp-1];
		stoff = val.u.i;
		val2 = tointeger(stack[sp-2]);
		idx = val2.u.i;
		if ((0 <= idx) && (idx < stack[stoff].u.a->ndim2))
			val = stack[stoff].u.a->ptr[idx];
		else {
			val.type = INTEGER;
			val.u.i=0;
		}
		push(val);
		break;
	case LVALVECCOPY2:
		val = stack[sp-1];
		stoff = val.u.i;
		val2 = tointeger(stack[sp-2]);
		idx = val2.u.i;
		val2 = tointeger(stack[sp-3]);
		idx += val2.u.i*stack[stoff].u.a->ndim2;
		if ((0 <= idx) && (idx < (stack[stoff].u.a->ndim1*stack[stoff].u.a->ndim2)))
			val = stack[stoff].u.a->ptr[idx];
		else {
			val.type = INTEGER;
			val.u.i=0;
		}
		push(val);
		break;
	case LVALVARSTORE:
	case LVALNODESTORE:
		val = pop();
		val2 = pop();
		stoff = val2.u.i;
		stack[stoff] = val;
		break;
	case LVALVECSTORE1:
	case LVALNODEVECSTORE1:
		val = pop();
		val2 = pop();
		stoff = val2.u.i;
		val2 = tointeger(pop());
		idx = val2.u.i;
		if ((0 <= idx) && (idx < stack[stoff].u.a->ndim2))
			stack[stoff].u.a->ptr[idx] = val;
		break;
	case LVALVECSTORE2:
	case LVALNODEVECSTORE2:
		val = pop();
		val2 = pop();
		stoff = val2.u.i;
		val2 = tointeger(pop());
		idx = val2.u.i;
		val2 = tointeger(pop());
		idx += val2.u.i*stack[stoff].u.a->ndim2;
#ifdef IAG
printf("***PUT idx=%d ndim1=%d ndim2=%d\n",idx,stack[stoff].u.a->ndim1,stack[stoff].u.a->ndim2);
#endif
		if ((0 <= idx) && (idx < (stack[stoff].u.a->ndim1*stack[stoff].u.a->ndim2)))
			stack[stoff].u.a->ptr[idx] = val;
		break;
	case LVALVECINIT:
		idx = code[pc++];
		val2 = pop();
		stoff = val2.u.i;
		while (idx-- > 0) {
			val = pop();
			if ((0 <= idx) && (idx < (stack[stoff].u.a->ndim1*stack[stoff].u.a->ndim2)))
				stack[stoff].u.a->ptr[idx] = val;
		}
		break;
	case ADD:
		val = tonumber(pop());
		val2 = tonumber(pop());
		if ((val.type==INTEGER) && (val2.type==INTEGER)) {
			val.u.i += val2.u.i;
			push(val);
		}
		else {
			val = tofloat(val);
			val2 = tofloat(val2);
			val.u.d += val2.u.d;
			push(val);
		}
		break;
	case SUBTRACT:
		val2 = tonumber(pop());
		val = tonumber(pop());
		if ((val.type==INTEGER) && (val2.type==INTEGER)) {
			val.u.i -= val2.u.i;
			push(val);
		}
		else {
			val = tofloat(val);
			val2 = tofloat(val2);
			val.u.d -= val2.u.d;
			push(val);
		}
		break;
	case MULTIPLY:
		val = tonumber(pop());
		val2 = tonumber(pop());
		if ((val.type==INTEGER) && (val2.type==INTEGER)) {
			val.u.i *= val2.u.i;
			push(val);
		}
		else {
			val = tofloat(val);
			val2 = tofloat(val2);
			val.u.d *= val2.u.d;
			push(val);
		}
		break;
	case DIVIDE:
		val2 = tonumber(pop());
		val = tonumber(pop());
		if ((val.type==INTEGER) && (val2.type==INTEGER)) {
			if (val2.u.i==0) {
				fprintf(stderr,"Warning: division by zero error at PC:%04d\n",globalpc);
				val2.u.i=1;
			}
			val.u.i /= val2.u.i;
			push(val);
		}
		else {
			val = tofloat(val);
			val2 = tofloat(val2);
			if (val2.u.d==0) {
				fprintf(stderr,"Warning: division by zero error at PC:%04d\n",globalpc);
				val2.u.d=1;
			}
			val.u.d /= val2.u.d;
			push(val);
		}
		break;
	case NEGATE:
		val = tonumber(pop());
		if (val.type==FLOAT)
			val.u.d = -val.u.d;
		else
			val.u.i = -val.u.i;
		push(val);
		break;
	case CONCATENATE:
		val2 = tostring(pop());
		val = tostring(pop());
		s = malloc(strlen(val.u.s)+strlen(val2.u.s)+1);
		strcpy(s,val.u.s);
		strcat(s,val2.u.s);
		val.u.s = strsave(s);
		push(val);
		break;
	case IF:
		pc = perform(item,dct,pc,CONDEND);
		break;
	case JZ:
		val = pop();
		addr = code[pc++];
		if (val.u.i==0) pc = addr;
		break;
	case JNZ:
		val = pop();
		addr = code[pc++];
		if (val.u.i!=0) pc = addr;
		break;
	case JMP:
		addr = code[pc++];
		pc = addr;
		break;
	case TESTEQ:
		val2 = pop();
		val = pop();
		val.u.i = (valcompare(val,val2)==0) ? 1 : 0;
		val.type = INTEGER;
		push(val);
		break;
	case TESTNE:
		val2 = pop();
		val = pop();
		val.u.i = (valcompare(val,val2)!=0) ? 1 : 0;
		val.type = INTEGER;
		push(val);
		break;
	case TESTLT:
		val2 = pop();
		val = pop();
		val.u.i = (valcompare(val,val2)<0) ? 1 : 0;
		val.type = INTEGER;
		push(val);
		break;
	case TESTLE:
		val2 = pop();
		val = pop();
		val.u.i = (valcompare(val,val2)<=0) ? 1 : 0;
		val.type = INTEGER;
		push(val);
		break;
	case TESTGT:
		val2 = pop();
		val = pop();
		val.u.i = (valcompare(val,val2)>0) ? 1 : 0;
		val.type = INTEGER;
		push(val);
		break;
	case TESTGE:
		val2 = pop();
		val = pop();
		val.u.i = (valcompare(val,val2)>=0) ? 1 : 0;
		val.type = INTEGER;
		push(val);
		break;
	case LOGICALOR:
		val2 = tointeger(pop());
		val = tointeger(pop());
		if ((val.u.i!=0)||(val2.u.i!=0)) val.u.i=1;
		push(val);
		break;
	case LOGICALAND:
		val2 = tointeger(pop());
		val = tointeger(pop());
		if ((val.u.i==0)||(val2.u.i==0)) val.u.i=0;
		push(val);
		break;
	case LOGICALNOT:
		val = tointeger(pop());
		if (val.u.i==0) val.u.i=1; else val.u.i=0;
		push(val);
		break;
	case GSTACKRESERVE:
		stoff = code[pc++];	/* get size of variables */
		sp += stoff;		/* jump stack pointer over locals */
		if (sp > MAXSTACK)
			prxerror("stack overflow - too many global variables","");
		break;
	case STACKRESERVE:
		val.type = INTEGER;
		val.u.i = fnoff;
		push(val);		/* save current frame ptr */
		fnoff = sp;		/* set frame pointer to stop of stack */
		stoff = code[pc++];	/* get size of locals */
		sp += stoff;		/* jump stack pointer over locals */
		if (sp > MAXSTACK)
			prxerror("stack overflow - too many automatic variables","");
		break;
	case STACKFREE:
		sp = fnoff;		/* restore stack pointer */
		val = pop();		/* restore frame pointer */
		fnoff = val.u.i;
		break;
	case ASCEND:
		node = pop();
		if ((node.type==NODE)&&(node.u.v!=NULL)) {
			data = ((NSL_Item *)node.u.v)->in;
			if (!data || !data->in)
				/* no parent */
				node.u.v = NULL;
			else
				/* found parent */
				node.u.v = data->in;
		}
		push(node);
		break;
	case DESCEND:
		val = tostring(pop());
		node = pop();
		if ((node.type==NODE)&&(node.u.v!=NULL)) {
			for (data = ((NSL_Item *)node.u.v)->data;data!=NULL;data=data->next) {
				if (data->type==NSL_item_data) {
					itemp = data->first;
					if ((*val.u.s=='\0') || (strcmp(itemp->label,val.u.s)==0)) {
						/* got it ! */
						node.u.v = itemp;
						break;
					}
				}
				else if (data->type==NSL_text_data) {
					if (*val.u.s=='\0') {
						/* got one ! */
						node.type = STRING;
						node.u.s = strsave(data->first);
						break;
					}
				}
			}
			if (data==NULL)
				/* failed to find descendant */
				node.u.v = NULL;
		}
		push(node);
		break;
	case DESCENDSEL:
		val2 = tointeger(pop());
		val = tostring(pop());
		node = pop();
		if ((node.type==NODE)&&(node.u.v!=NULL)) {
			for (data = ((NSL_Item *)node.u.v)->data;data!=NULL;data=data->next) {
				if (data->type==NSL_item_data) {
					itemp = data->first;
					if ((*val.u.s=='\0') || (strcmp(itemp->label,val.u.s)==0)) {
						/* got one ! */
						val2.u.i--;
						if (val2.u.i <= 0) {
							node.type = NODE;
							node.u.v = itemp;
							break;
						}
					}
				}
				else if (data->type==NSL_text_data) {
					if (*val.u.s=='\0') {
						/* got one ! */
						val2.u.i--;
						if (val2.u.i <= 0) {
							node.type = STRING;
							node.u.s = strsave(data->first);
							break;
						}
					}
				}
			}
			if ((data==NULL) || (val2.u.i!=0)) {
				/* failed to find descendant */
				node.type = NODE;
				node.u.v = NULL;
			}
		}
		push(node);
		break;
	case DESCENDALL:
		val = tostring(pop());
		node = pop();
		pcexit = pc+1;
		if ((node.type==NODE)&&(node.u.v!=NULL)) {
			for (data = ((NSL_Item *)node.u.v)->data;data!=NULL;data=data->next) {
				if (data->type==NSL_item_data) {
					itemp = data->first;
#ifdef IAG
printf("In DESCENDALL matching %s against %s\n",val.u.s,itemp->label);
#endif
					if ((*val.u.s=='\0') || (strcmp(itemp->label,val.u.s)==0)) {
#ifdef IAG
printf("DESCENDALL matched %s\n",itemp->label);
#endif
						/* a match, absorb it */
						node.type = NODE;
						node.u.v = itemp;
						push(node);
						pcexit = perform(item,dct,pc,PATTEND);
						
						/* now try for more */
						node.type = NODE;
						node.u.v = itemp;
						push(node);
						push(val);
						pcexit = perform(item,dct,pc-1,PATTEND);
					}
					else {
#ifdef IAG
printf("DESCENDALL failed on %s\n",itemp->label);
#endif
						/* no match, try again at next level down */
						node.type = NODE;
						node.u.v = itemp;
						push(node);
						push(val);
						pcexit = perform(item,dct,pc-1,PATTEND);
					}
				}
				else if (data->type==NSL_text_data) {
					if (*val.u.s=='\0') {
						/* a match, absorb it */
						node.type = STRING;
						node.u.s = strsave(data->first);
						push(node);
						pcexit = perform(item,dct,pc,PATTEND);
					}
				}
			}
		}
		node.type = NODE;
		node.u.v=NULL;
		push(node);
		pc = pcexit-1;
		break;
	case DESCENDREC:
		node = pop();
		pcexit = pc+1;
		if ((node.type==NODE)&&(node.u.v!=NULL)) {
			for (data = ((NSL_Item *)node.u.v)->data;data!=NULL;data=data->next) {
				if (data->type==NSL_item_data) {
					itemp = data->first;
					node.u.v = itemp;
					push(node);
					/* process this node */
					pcexit = perform(item,dct,pc,PATTEND);
					/* process the children of this node */
					push(node);
					pcexit = perform(item,dct,pc-1,PATTEND);
				}
			}
		}
		node.u.v=NULL;
		push(node);
		pc = pcexit-1;
		break;
	case FUNCTIONCALL:
		entry = code[pc++];
		if (symtab[entry]->status==RESERVED) {
			/* built in function */
			val = symtab[entry]->val;
			(*(void (*)(void))(val.u.v))();
		}
		else {
			/* clear out return value */
			gretval.type=INTEGER;
			gretval.u.i = 0;
			/* call user function */
			perform(item,dct,symtab[entry]->val.u.i,FUNCEND);
			/* remove parameters from stack */
			for (i=0;i<symtab[entry]->nparm;i++) pop();
			/* push return value on stack */
			push(gretval);
		}
		break;
	case RETURN:
		/* save return expression */
		gretval = pop();
		break;
	case NTESTEQ:
		node2 = pop();
		node = pop();
		val.u.i = (node.u.v==node2.u.v) ? 1 : 0;
		val.type = INTEGER;
		push(val);
		break;
	case NTESTNE:
		node2 = pop();
		node = pop();
		val.u.i = (node.u.v!=node2.u.v) ? 1 : 0;
		val.type = INTEGER;
		push(val);
		break;
	case POP:
		val = pop();
		break;
	case FOREACH:
		stoff = fnoff + code[pc++];
		pcexit = code[pc++];
		pccode = code[pc++];
		/* generate list of matching items */
		sp1 = sp;
		perform(item,dct,pc,PATTEND);
		sp2 = sp;
#ifdef IAG
printf("FOREACH found %d items\n",sp2-sp1);
#endif
		/* now run code block */
		for (i=sp1;i<sp2;i++) {
			stack[stoff] = stack[i];
			if (stack[stoff].u.v!=NULL)
				perform(item,dct,pccode,FUNCEND);
		}
		/* restore stack */
		sp = sp1;
		/* jump to end */
		pc = pcexit;
		break;
	case GFOREACH:
		stoff = code[pc++];
		pcexit = code[pc++];
		pccode = code[pc++];
		/* generate list of matching items */
		sp1 = sp;
		perform(item,dct,pc,PATTEND);
		sp2 = sp;
		/* now run code block */
		for (i=sp1;i<sp2;i++) {
			stack[stoff] = stack[i];
			if (stack[stoff].u.v!=NULL)
				perform(item,dct,pccode,FUNCEND);
		}
		/* restore stack */
		sp = sp1;
		/* jump to end */
		pc = pcexit;
		break;
	case PATTSAVE:
		node = pop();
		if ((node.type==NODE)&&(node.u.v!=NULL)) {
itemp = (NSL_Item *)node.u.v;
#ifdef IAG
printf("PATTSAVE saving <%s>\n",itemp->label);
#endif
			push(node);
		}
		else if ((node.type==STRING)&&(node.u.s!=NULL)) {
#ifdef IAG
printf("PATTSAVE saving '%s'\n",node.u.s);
#endif
			push(node);
		}
#if 0
		else {
			push(node);
		}
#endif
		break;
	case PATTEND:
		break;
	case DUPSTACK2:
		val2 = pop();
		val = pop();
		push(val);
		push(val2);
		push(val);
		push(val2);
		break;
	case SIZEOF:
		stoff = fnoff + code[pc++];
		val.u.i = (stack[stoff].u.a->ndim1 * stack[stoff].u.a->ndim2);
		val.type = INTEGER;
		push(val);
		break;
	case GSIZEOF:
		stoff = code[pc++];
		val.u.i = (stack[stoff].u.a->ndim1 * stack[stoff].u.a->ndim2);
		val.type = INTEGER;
		push(val);
		break;
	case ARRSIZE:
		stoff = fnoff + code[pc++];
		val = tointeger(pop());
		if (val.u.i >= 1) {
			stack[stoff].u.a->ndim1=1;
			stack[stoff].u.a->ndim2=val.u.i;
			stack[stoff].u.a->ptr=(struct Value_rec *)realloc(stack[stoff].u.a->ptr,val.u.i*sizeof(struct Value_rec));
		}
		break;
	case GARRSIZE:
		stoff = code[pc++];
		val = tointeger(pop());
		if (val.u.i >= 1) {
			stack[stoff].u.a->ndim1=1;
			stack[stoff].u.a->ndim2=val.u.i;
			stack[stoff].u.a->ptr=(struct Value_rec *)realloc(stack[stoff].u.a->ptr,val.u.i*sizeof(struct Value_rec));
		}
		break;
	case ARRSIZE2:
		stoff = fnoff + code[pc++];
		val2 = tointeger(pop());
		val = tointeger(pop());
		if ((val.u.i>=1)&&(val2.u.i>=1)) {
			stack[stoff].u.a->ndim1=val.u.i;
			stack[stoff].u.a->ndim2=val2.u.i;
			stack[stoff].u.a->ptr=(struct Value_rec *)realloc(stack[stoff].u.a->ptr,val.u.i*val2.u.i*sizeof(struct Value_rec));
		}
		break;
	case GARRSIZE2:
		stoff = code[pc++];
		val2 = tointeger(pop());
		val = tointeger(pop());
		if ((val.u.i>=1)&&(val2.u.i>=1)) {
			stack[stoff].u.a->ndim1=val.u.i;
			stack[stoff].u.a->ndim2=val2.u.i;
			stack[stoff].u.a->ptr=(struct Value_rec *)realloc(stack[stoff].u.a->ptr,val.u.i*val2.u.i*sizeof(struct Value_rec));
		}
		break;
	case ASSERT:
		stoff = code[pc++];
		val = tointeger(pop());
		if (val.u.i==0) {
			fprintf(stderr,"Assert fails at line %d (PC:%04d)\n",stoff,globalpc);
			exit(1);
		}
		break;
	case ASSERTVAL:
		stoff = code[pc++];
		val2 = tostring(pop());
		val = tointeger(pop());
		if (val.u.i==0) {
			fprintf(stderr,"Assert fails at line %d (PC:%04d): %s\n",stoff,globalpc,val2.u.s);
			exit(1);
		}
		break;
	default:
		fprintf(stderr,"Panic! op=%d pc=%d\n",op,pc);
		prxerror("Unknown op-code in interpreter!","");
		break;
	}
	}
	return(pc);
}

/* run function with no context */
int run_empty(NSL_Doctype dct,char *fname)
{
	int	i;
	struct function_pc *fpc;
	for (i=0;i<funcpos;i++) {
		if (_stricmp(functab[i].name,fname)==0) {
			fpc = functab[i].first;
			while (fpc) {
				perform(NULL,dct,fpc->pc,FUNCEND);
				fpc = fpc->next;
			}
		}
	}
	return(0);
}

/* run function on root context */
int run_root(NSL_Item *item,NSL_Doctype dct,char *fname)
{
	int		i;
	struct function_pc *fpc;

	/* process this item */
	for (i=0;i<funcpos;i++) {
		if (_stricmp(functab[i].name,fname)==0) {
			fpc = functab[i].first;
			while (fpc) {
				perform(item,dct,fpc->pc,FUNCEND);
				fpc = fpc->next;
			}
		}
	}
	return(0);
}

/* run function on all nodes */
int run_all(NSL_Item *item,NSL_Doctype dct,char *fname)
{
	int		i;
	NSL_Data	*data;
	struct function_pc *fpc;

	/* process this item */
	for (i=0;i<funcpos;i++) {
		if (_stricmp(functab[i].name,fname)==0) {
			fpc = functab[i].first;
			while (fpc) {
				perform(item,dct,fpc->pc,FUNCEND);
				fpc = fpc->next;
			}
		}
	}

	/* then process all decendants */	
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			run_all(item,dct,fname);
		}
	}
	return(0);
}

/* run named functions on all nodes */
int run_named(NSL_Item *item,NSL_Doctype dct,int level)
{
	int		i;
	NSL_Data	*data;
	struct function_pc *fpc;

	/* process this item */
	for (i=0;i<funcpos;i++) {
		if ((_stricmp(item->label,functab[i].name)==0) ||
		    (strcmp(".",functab[i].name)==0) ||
		    ((strcmp("/",functab[i].name)==0)&&(level==0))) {
#ifdef IAG
printf("Running function %d of %d called '%s'\n",i,funcpos,functab[i].name);
#endif
			fpc = functab[i].first;
			while (fpc) {
				perform(item,dct,fpc->pc,FUNCEND);
				fpc = fpc->next;
			}
		}
	}

	/* then process all decendants */	
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			run_named(item,dct,level+1);
		}
	}

	return(0);
}

/* run program against a given XML tree */
int run(NSL_Item *item,NSL_Doctype dct)
{
	/* run initialisation */
	run_empty(dct,"!start");

	/* run root initialisation */
	run_root(item,dct,"/start");

	/* run node initialisation */
	run_all(item,dct,".start");

	/* run named nodes */
	run_named(item,dct,0);

	/* run node termination */
	run_all(item,dct,".end");

	/* run root termination */
	run_root(item,dct,"/end");

	/* run termination */
	run_empty(dct,"!end");
		
	return(0);
}

/* global doctype */
NSL_Doctype 	gdct=NULL;

/* process - process XML file using compiled code */
int process(char *xfilename,char *ofilename)
{
	NSL_Bit		*bit;
	NSL_Item	*item;
	NSL_File 	inf,outf=NULL;
	int		first=1;
	char		tfilename[256];

	/* Initialise the NSL API */
	NSLInit(0);
	groot=NULL;

	/* open input file */
	if (xfilename && *xfilename && *xfilename!='-') {
		if ((inf = SFopen(xfilename, gdct, NSL_read))==NULL)
			prxerror("unable to open '%s'",xfilename);
	}
	else if (xfilename && *xfilename && *xfilename=='-') {
		inf = SFFopen(stdin, gdct, NSL_read, "stdin");
	}
	else {
		/* initialise global variables */
		perform(NULL,gdct,0,END);
		return(0);
	}
	gdct = DoctypeFromFile(inf);

	/* initialise global variables in script */
	perform(NULL,gdct,0,END);

	/* get output filename */
	if (ofilename && *ofilename && *ofilename!='-') {
		strcpy(tfilename,ofilename);
		strcat(tfilename,".tmp");
	}
	else
		tfilename[0]='\0';

	/* read first tag */
	while ((bit=GetNextBit(inf))!=NULL) {
		if (bit->type!=NSL_start_bit) 
			prxerror("Could not read start tag","");

		/* load the whole thing into memory */
		item = ItemParse(inf,bit->value.item);

		/* run the program on it */
		groot=item;
		run(item,gdct);

		/* print the modified item */
		if (first) {
			if (!ofilename)
				outf=NULL;
			else if (*ofilename && *ofilename!='-') {
				if ((outf = SFopen(tfilename, gdct, NSL_write_pretty))==NULL)
					prxerror("unable to open '%s'",tfilename);
			}
			else {
				/* Open the output stream to standard output */
				outf = SFFopen(stdout, gdct, NSL_write_pretty, "stdout");
			}
			first=0;
		}
	
		if (outf) {
			PrintItem(outf,item);
			ForceNewline(outf);
		}

	}

	/* that's all folks */
	if (outf) SFclose(outf);
	SFclose(inf);
	if (tfilename[0]) {
		remove(ofilename);
		rename(tfilename,ofilename);
	}

	return(0);
}

