/* process.h -- definitions for XML processing interpreter */

/* interpreter instructions */
enum CODE {
	ERROR=0,	/* illegal value */
	END,		/* end of program */
	FUNCSTART,	/* start of a function */
	FUNCEND,	/* end of a function */
	VARSTORE,	/* store value in variable */
	VARLOAD,	/* load value from variable */
	GVARSTORE,	/* store value in variable */
	GVARLOAD,	/* load value from variable */
	VECINIT1,	/* initialise local 1D array */
	GVECINIT1,	/* initialise global 1D array */
	VELLOAD1,	/* load value from 1D array */
	GVELLOAD1,	/* load value from 1D array */
	VECINIT2,	/* initialise local 2D array */
	GVECINIT2,	/* initialise global 2D array */
	VELLOAD2,	/* load value from 2D array */
	GVELLOAD2,	/* load value from 2D array */
	NODELOADROOT,	/* load node from root item */
	NODELOADTHIS,	/* load node from current item */
	NODELOADVAR,	/* load node from variable */
	GNODELOADVAR,	/* load node from variable */
	NODELOADANY,	/* load 'match any' node */
	NODELOADNIL,	/* load nil node */
	NODELOAD,	/* load node from symbol table */
	NODESTORE,	/* store node from stack */
	GNODESTORE,	/* store node from stack */
	ATTRSTORE,	/* store value in attribute */
	ATTRLOAD,	/* load value from attribute */
	CONLOAD,	/* load constant value */
	NAMELOAD,	/* load identifier name */
	PUSHLVAL,
	GPUSHLVAL,
	LVALVARCOPY,
	LVALVECCOPY1,
	LVALVECCOPY2,
	LVALVARSTORE,
	LVALVECSTORE1,
	LVALVECSTORE2,
	LVALVECINIT,
	LVALNODESTORE,
	LVALNODEVECSTORE1,
	LVALNODEVECSTORE2,
	ADD,
	SUBTRACT,
	MULTIPLY,
	DIVIDE,
	NEGATE,
	CONCATENATE,
	IF,		/* conditional statement */
	CONDEND,	/* condition end */
	JZ,		/* jump if zero */
	JNZ,		/* jump if non zero */
	JMP,		/* unconditional jump */
	TESTEQ,
	TESTNE,
	TESTLT,
	TESTLE,
	TESTGT,
	TESTGE,
	LOGICALOR,
	LOGICALAND,
	LOGICALNOT,
	GSTACKRESERVE,
	STACKRESERVE,
	STACKFREE,
	ASCEND,
	DESCEND,
	DESCENDSEL,
	FUNCTIONCALL,
	RETURN,
	NTESTEQ,
	NTESTNE,
	POP,
	FOREACH,
	GFOREACH,
	DESCENDEACH,
	DESCENDALL,
	DESCENDREC,
	PATTSAVE,
	PATTEND,
	DUPSTACK2,
	SIZEOF,
	GSIZEOF,
	ARRSIZE,
	GARRSIZE,
	ARRSIZE2,
	GARRSIZE2,
	LOADTRUE,
	LOADFALSE,
	ASSERT,
	ASSERTVAL
};

int *code;		/* compiled code */
int codepos;		/* current position */
int codesize;		/* maximum allocation */
#define CODECHUNKSIZE	4096

/* routines in process.c */
int codeexpand(int v);
void dumpcode();
int addfunction(char *name,int pc);
int  process(char *filename,char *ofilename);
void push(Value val);
Value pop();
Value tochar(Value val);
Value tointeger(Value val);
Value tofloat(Value val);
Value tostring(Value val);
int isinteger(Value val);
Value tonumber(Value val);
int valcompare(Value val1,Value val2);
int perform(NSL_Item *item,NSL_Doctype dct,int pc,int final);


#ifdef IAG
#define EMIT1(v) ( ((fprintf(stderr,"\nCODE=%d\n",(int)(v))) && (codepos < codesize)) ? code[codepos++]=(int)(v) : codeexpand(v) )
#else
#define EMIT1(v) ( (codepos < codesize) ? code[codepos++]=(int)(v) : codeexpand(v) )
#endif
#define EMIT2(v1,v2) { EMIT1(v1); EMIT1(v2); }


/* current function offset into stack */
int fnoff;

/* global function return value */
Value gretval;

/* root address */
NSL_Item *groot;

/* global doctype */
NSL_Doctype 	gdct;

