/* prx.h - header file for ProXML */

#define YYDEBUG 1

/* Mark Huckvale - University College London */

/* version 1.1 - July 2000 */

#define PROGNAME "prx"
#define PROGVERS "1.31"

/* array */
struct array_rec {
	int	ndim1;
	int	ndim2;
	struct Value_rec *ptr;
};

/* constant value */
typedef struct Value_rec {
	short	type;
	union {
		char	c;
		int	i;
		double	d;
		char	*s;
		void	*v;
		struct array_rec *a;
	} u;
} Value;

/* symbol table */
typedef struct Symrec {
	char	*name;		/* symbol name */
	int	type;		/* symbol type */
	int	status;		/* symbol status */
	int	stackpos;	/* stack position */
	Value	val;		/* value */
	int	etype;		/* expression type for functions */
	int	nparm;		/* # parameters */
} Symbol;

/* symbol table */
#define MAXSYMBOL 4000
Symbol 	*symtab[MAXSYMBOL];
int	sympos;

/* values for symbol status */
#define RESERVED   0
#define UNDECLARED 1
#define ADRPAR     2
#define VALPAR     3
#define LOCAL      4
#define GLOBAL     5
#define BLTIN      6
#define CONSTANT   7
#define VARPARAM   0x40

/* compiler symbols */
#include "prxtab.h"

/* prototypes */
void yyerror(char *s);
int parsefail(char *s,YYLTYPE *yylloc);
int warning(YYLTYPE *yylloc,char *s,char *t);
void prxerror(char *s,char *t);

/* spclex.c */
char *strsave(const char *str);
int installsymbol(char *name,int type,int status,char *valstr);
void freesymbol(Symbol *sym);
void clearlocals(int epos);
void initsymbol(void);
int lookupsymbol(char *name);
void resetlex(void);
char *findfile(char *name);
void pushfcontext(char *fname);
int popfcontext(void);
int charget(void);
void charunget(int c);
int yylex(void);
int backslash(int c);
int follow(int c2,int expect,int ifyes,int ifno);
int assignlocal(int sym,int vno);
int assignnode(int sym,int vno);
int assignlocalfunc(int sym,int vno);
int assignnodefunc(int sym,int vno);
int assignglobal(int sym,int vno);
int assignglobalnode(int sym,int vno);
int assignglobalconstant(int sym,int vno);
int assignlocalvector(int sym,int vno);
int assignglobalvector(int sym,int vno);
int assignlocalnodevector(int sym,int vno);
int assignglobalnodevector(int sym,int vno);
int correctparams(int nparam);

