/* prx.y -- yacc description for ProXML */

/* version 2.0 - June 2000 */

%{
#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#else
int getopt(int,char **,char *);
#endif
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <malloc.h>
#include <time.h>
#ifdef WIN32
#include <process.h>
#define alloca _alloca
#endif
#ifdef SVR4
#define alloca malloc
#endif
#include "lt-memory.h"
#include "nsl.h"
#include "prx.h"
#include "process.h"

/* global decoding states */
int	parno;		/* parameter # in function */
int	varno;		/* variable # in function */
int	gvarno;		/* global variable # */
int	infunc;		/* 1=varfunc, 2=nodefunc, 3=elementfunc */
int	tmp;

/* context control */
int	pcstack[64];
int	pcsp;
void pcpush(int pc) { pcstack[pcsp++]=pc; }
int pcpop() { return pcstack[--pcsp]; }
void pcswap() { int i1=pcpop(); int i2=pcpop(); pcpush(i1); pcpush(i2); }

%}

/* ignore ambiguity introduced by if-then-else */
%expect 1

/* YACC token identification */
%token	ARRAY
%token	NARRAY
%token  ASSERT_TOK
%token	BREAK_TOK
%token	CHAR
%token	CONSTSYM
%token	CONST_TOK
%token	DESCENDANT_TOK
%token	ELSE_TOK
%token  FALSE_TOK
%token	FLOAT
%token	FOR_TOK
%token	FOREACH_TOK
%token	FUNCTION
%token	IDENTIFIER
%token	IF_TOK
%token	INCLUDE_TOK
%token	INTEGER
%token	NIL_TOK
%token	NODE
%token	NODE_FUNCTION
%token	NODE_TOK
%token	PARENT_TOK
%token	RESIZE_TOK
%token	RETURN_TOK
%token	SIZEOF_TOK
%token	STRING
%token	TRUE_TOK
%token	VALUE
%token	VARIABLE
%token	VAR_TOK
%token	WHILE_TOK

/* arithmetic expressions */
%right	ASSIGN_TOK
%right	PLUS_ASSIGN_TOK
%right	TIMES_ASSIGN_TOK
%right	'.'
%left	OR_TOK BIT_OR_TOK
%left	AND_TOK BIT_AND_TOK
%left	GT_TOK GE_TOK LT_TOK LE_TOK EQ_TOK NE_TOK
%left	CONCAT_TOK
%left	PLUS_TOK MINUS_TOK
%left	TIMES_TOK DIVIDE_TOK
%left	UNARYMINUS NOT_TOK

/* trying to define a program */
%start Program
%%

/* Definition of ProXML grammar */
Program:
		ComponentList		{ EMIT1(END); }
		;

ComponentList:
		/* empty */
		|
		Component
		ComponentList
		;

Component:
		/* include file */
		INCLUDE_TOK
		STRING			{ pushfcontext(symtab[$2]->val.u.s); }
		';'
		|
		/* ELEMENT function */
		IdentifierList		{ EMIT1(JMP);
					  pcpush(codepos);
					  EMIT1(ERROR);
					  EMIT2(FUNCSTART,$1); 
					  pcpush(sympos);
					  varno=0;
					  infunc=3;
					  EMIT1(STACKRESERVE);
					  pcpush(codepos);
					  EMIT1(0);
					}
		CompoundStatement	{ code[pcpop()] = varno;
					  EMIT1(STACKFREE);
					  EMIT1(FUNCEND); 
					  clearlocals(pcpop());
					  code[pcpop()] = codepos;
					  infunc=0;
					}
		|
		/* user function */
		VarToken
		IDENTIFIER
		'('			{ code[pcpop()] = varno;
					  EMIT1(JMP);
					  pcpush(codepos);
					  EMIT1(ERROR);
					  pcpush(sympos);
					  parno=0; 
					  infunc=1;
					}
		DummyParameterList
		')'			{ EMIT2(FUNCSTART,$2); 
					  symtab[$2]->type = FUNCTION;
					  symtab[$2]->status = GLOBAL;
					  symtab[$2]->val.type = INTEGER;
					  symtab[$2]->val.u.i = codepos;
					  symtab[$2]->nparm = parno;
					  correctparams(parno);
					  varno=0;
					  EMIT1(STACKRESERVE);
					  pcpush(codepos);
					  EMIT1(0);
					}
		CompoundStatement	{ code[pcpop()] = varno;
					  EMIT1(STACKFREE);
					  EMIT1(FUNCEND); 
					  clearlocals(pcpop());
					  code[pcpop()] = codepos;
					  infunc=0;
					}
		|
		/* user node function */
		NodeToken
		IDENTIFIER
		'('			{ code[pcpop()] = varno;
					  EMIT1(JMP);
					  pcpush(codepos);
					  EMIT1(ERROR);
					  pcpush(sympos);
					  parno=0; 
					  infunc=2;
					}
		DummyParameterList
		')'			{ EMIT2(FUNCSTART,$2); 
					  symtab[$2]->type = NODE_FUNCTION;
					  symtab[$2]->status = GLOBAL;
					  symtab[$2]->val.type = INTEGER;
					  symtab[$2]->val.u.i = codepos;
					  symtab[$2]->nparm = parno;
					  correctparams(parno);
					  varno=0;
					  EMIT1(STACKRESERVE);
					  pcpush(codepos);
					  EMIT1(0);
					}
		CompoundStatement	{ code[pcpop()] = varno;
					  EMIT1(STACKFREE);
					  EMIT1(FUNCEND); 
					  clearlocals(pcpop());
					  code[pcpop()] = codepos;
					  infunc=0;
					}
		|
		/* global node variables */
		NodeToken
		GlobalNode
		GlobalNodeCont
		';'			{ code[pcpop()] = varno;
					}
		|
		/* global constants */
		CONST_TOK
		IDENTIFIER		{ varno=0;
					  EMIT1(GSTACKRESERVE);
					  pcpush(codepos);
					  EMIT1(0);
					}
		ASSIGN_TOK
		Expression		{ EMIT2(GVARSTORE,gvarno); 
					  assignglobalconstant($2,gvarno++);
					  varno++;
					}
		GlobalConstantContinue
		';'			{ code[pcpop()] = varno;
					}
		|
		/* global variables */
		VarToken
		GlobalVariable
		GlobalVariableCont
		';'			{ code[pcpop()] = varno;
					}
		;

VarToken:
		VAR_TOK			{ varno=0;
					  EMIT1(GSTACKRESERVE);
					  pcpush(codepos);
					  EMIT1(0);
					}
		;
NodeToken:
		NODE_TOK		{ varno=0;
					  EMIT1(GSTACKRESERVE);
					  pcpush(codepos);
					  EMIT1(0);
					}
		;

GlobalVariableCont:
		/* empty */
		|
		','
		GlobalVariable
		GlobalVariableCont
		;

GlobalVariable:
		/* global variable */
		IDENTIFIER		
		VarInitialisation	{ if ($2==1) EMIT2(GVARSTORE,gvarno); 
					  assignglobal($1,gvarno++);
					  varno++;
					}
		|
		/* global vector */
		IDENTIFIER
		'['
		Expression
		']'			{ EMIT2(GVECINIT1,gvarno);
					  assignglobalvector($1,gvarno++);
					  varno++;
					  parno=0;
					}
		VecInitialisation	{ if ($6==1) {
						EMIT2(GPUSHLVAL,symtab[$1]->stackpos);
						EMIT2(LVALVECINIT,parno);
					  }
					}
		|
		/* global vector */
		IDENTIFIER
		'['
		Expression
		','
		Expression
		']'			{ EMIT2(GVECINIT2,gvarno);
					  assignglobalvector($1,gvarno++);
					  varno++;
					  parno=0;
					}
		VecInitialisation	{ if ($8==1) {
						EMIT2(GPUSHLVAL,symtab[$1]->stackpos);
						EMIT2(LVALVECINIT,parno);
					  }
					}
		;

GlobalNodeCont:
		/* empty */
		|
		','
		GlobalNode
		GlobalNodeCont
		;

GlobalNode:
		/* global node */
		IDENTIFIER
		NodeVarInitialisation	
					{ if ($2==1) EMIT2(GVARSTORE,gvarno); 
					  assignglobalnode($1,gvarno++);
					  varno++;
					}
		|
		/* global node vector */
		IDENTIFIER
		'['
		Expression
		']'			{ EMIT2(GVECINIT1,gvarno);
					  assignglobalnodevector($1,gvarno++);
					  varno++;
					}
		|
		/* global vector */
		IDENTIFIER
		'['
		Expression
		','
		Expression
		']'			{ EMIT2(GVECINIT2,gvarno);
					  assignglobalnodevector($1,gvarno++);
					  varno++;
					  parno=0;
					}
		;


GlobalConstantContinue:
		/* empty */
		|
		','
		IDENTIFIER
		ASSIGN_TOK
		Expression			{ EMIT2(GVARSTORE,gvarno); 
						  assignglobalconstant($2,gvarno++);
						  varno++;
						}
		GlobalConstantContinue
		;

IdentifierList:
		Identifier		{ $$=$1; }
		|
		IdentifierList
		','
		Identifier		{ $$=$3; }
		;

Identifier:
		'.'			{ addfunction(".",codepos+4); $$=$1; }
		|
		DIVIDE_TOK		{ addfunction("/",codepos+4); $$=$1; }
		|
		'.'
		IDENTIFIER		{ char buff[256];
					  sprintf(buff,".%s",symtab[$2]->name);
					  addfunction(buff,codepos+4);
					  $$=$1;
					}
		|
		DIVIDE_TOK
		IDENTIFIER		{ char buff[256];
					  sprintf(buff,"/%s",symtab[$2]->name);
					  addfunction(buff,codepos+4);
					  $$=$1;
					}
		|
		NOT_TOK
		IDENTIFIER		{ char buff[256];
					  sprintf(buff,"!%s",symtab[$2]->name);
					  addfunction(buff,codepos+4);
					  $$=$1;
					}
		|
		IDENTIFIER		{ addfunction(symtab[$1]->name,codepos+4); $$=$1; }
		;

DummyParameterList:
		/* empty */
		|
		DummyParameterListOb
		;

DummyParameterListOb:
		DummyParameter
		|
		DummyParameterListOb
		','
		DummyParameter
		;

DummyParameter:
		VAR_TOK
		IDENTIFIER		{ assignlocal($2,++parno); }
		|
		NODE_TOK
		IDENTIFIER		{ assignnode($2,++parno); }
		|
		VAR_TOK
		IDENTIFIER
		'(' ')'			{ assignlocalfunc($2,++parno); }
		|
		NODE_TOK
		IDENTIFIER
		'(' ')'			{ assignnodefunc($2,++parno); }
		;

StatementList:
		/* empty */
		|
		SimpleOrCompoundStatement
		StatementList
		;

SimpleOrCompoundStatement:
		Statement
		|
		CompoundStatement
		;

CompoundStatement:
		'{'
		StatementList
		'}'
		;

Statement:
		AssignmentStatement
		';'
		|
		ProcedureStatement
		';'
		|
		IfStatement
		|
		ForStatement
		|
		WhileStatement
		|
		ForEachStatement
		|
		ReturnStatement
		';'
		|
		ResizeStatement
		';'
		|
		AssertStatement
		';'
		;

AssignmentStatement:
		DeclareStatement
		|
		VarAssignmentStatement
		|
		NodeAssignmentStatement
		|
		AttrAssignmentStatement
		;

DeclareStatement:
		VAR_TOK
		LocalVariable
		LocalVariableCont
		|
		NODE_TOK
		LocalNode
		LocalNodeCont
		;

LocalVariableCont:
		/* empty */
		|
		','
		LocalVariable
		LocalVariableCont
		;

LocalVariable:
		/* local variable */
		IDENTIFIER		
		VarInitialisation	{ if ($2==1) EMIT2(VARSTORE,varno); 
					  assignlocal($1,varno++);
					}
		|
		/* local vector */
		IDENTIFIER
		'['
		Expression
		']'			{ EMIT2(VECINIT1,varno);
					  assignlocalvector($1,varno++);
					  parno=0;
					}
		VecInitialisation	{ if ($6==1) {
						EMIT2(PUSHLVAL,symtab[$1]->stackpos);
						EMIT2(LVALVECINIT,parno);
					  }
					}
		|
		/* local 2D vector */
		IDENTIFIER
		'['
		Expression
		','
		Expression
		']'			{ EMIT2(VECINIT2,varno);
					  assignlocalvector($1,varno++);
					  parno=0;
					}
		VecInitialisation	{ if ($8==1) {
						EMIT2(GPUSHLVAL,symtab[$1]->stackpos);
						EMIT2(LVALVECINIT,parno);
					  }
					}
		;

VarInitialisation:
		/* empty */			{ $$=0; }
		|
		ASSIGN_TOK
		Expression			{ $$=1; }
		;

VecInitialisation:
		/* empty */			{ $$=0; }
		|
		ASSIGN_TOK
		'{'
		Expression			{ parno++; }
		VecInitCont
		'}'				{ $$=1; }
		;

VecInitCont:
		/* empty */
		|
		','
		Expression			{ parno++; }
		VecInitCont
		;

LocalNodeCont:
		/* empty */
		|
		','
		LocalNode
		LocalNodeCont
		;

LocalNode:
		/* local node */
		IDENTIFIER		
		NodeVarInitialisation	{ if ($2==1) EMIT2(NODESTORE,varno); 
					  assignnode($1,varno++);
					}
		|
		/* local node vector */
		IDENTIFIER
		'['
		Expression
		']'			{ EMIT2(VECINIT1,varno);
					  assignlocalnodevector($1,varno++);
					}
		|
		/* local node 2D vector */
		IDENTIFIER
		'['
		Expression
		','
		Expression
		']'			{ EMIT2(VECINIT2,varno);
					  assignlocalnodevector($1,varno++);
					}
		;

NodeVarInitialisation:
		/* empty */			{ $$=0; }
		|
		ASSIGN_TOK
		NodeExpression			{ $$=1; }
		;


VarAssignmentStatement:
		LValue
		ASSIGN_TOK
		Expression			{ if ($1==2)
							EMIT1(LVALVECSTORE2);
						  else if ($1==1)
							EMIT1(LVALVECSTORE1);
						  else
							EMIT1(LVALVARSTORE);
						}
		|
		LValue
		PLUS_ASSIGN_TOK			{ if ($1==2)
							EMIT1(LVALVECCOPY2); 
						  else if ($1==1)
							EMIT1(LVALVECCOPY1); 
						  else
							EMIT1(LVALVARCOPY);
						}
		Expression			{ EMIT1(ADD);
						  if ($1==2)
							EMIT1(LVALVECSTORE2);
						  else if ($1==1)
							EMIT1(LVALVECSTORE1);
						  else
							EMIT1(LVALVARSTORE);
						}
		|
		LValue
		TIMES_ASSIGN_TOK		{ if ($1==2)
							EMIT1(LVALVECCOPY2); 
						  else if ($1==1)
							EMIT1(LVALVECCOPY1); 
						  else
							EMIT1(LVALVARCOPY);
						}
		Expression			{ EMIT1(MULTIPLY);
						  if ($1==2)
							EMIT1(LVALVECSTORE2);
						  else if ($1==1)
							EMIT1(LVALVECSTORE1);
						  else
							EMIT1(LVALVARSTORE);
						}
		;

LValue:
		VARIABLE			{ if (symtab[$1]->status==GLOBAL) {
							EMIT2(GPUSHLVAL,symtab[$1]->stackpos);
						  }
						  else {
							EMIT2(PUSHLVAL,symtab[$1]->stackpos);
						  }
						  $$=0;
						}
		|
		ARRAY
		'['
		Expression
		']'				{ if (symtab[$1]->status==GLOBAL) {
							EMIT2(GPUSHLVAL,symtab[$1]->stackpos);
						  }
						  else {
							EMIT2(PUSHLVAL,symtab[$1]->stackpos);
						  }
						  $$=1;
						}
		|
		ARRAY
		'['
		Expression
		','
		Expression
		']'				{ if (symtab[$1]->status==GLOBAL) {
							EMIT2(GPUSHLVAL,symtab[$1]->stackpos);
						  }
						  else {
							EMIT2(PUSHLVAL,symtab[$1]->stackpos);
						  }
						  $$=2;
						}
		;

NodeAssignmentStatement:
		NodeLValue
		ASSIGN_TOK
		NodeExpression			{ if ($1==2)
							EMIT1(LVALNODEVECSTORE2);
						  else if ($1==1)
							EMIT1(LVALNODEVECSTORE1);
						  else
							EMIT1(LVALNODESTORE);
						}
		;

NodeLValue:
		NODE				{ if (symtab[$1]->status==GLOBAL) {
							EMIT2(GPUSHLVAL,symtab[$1]->stackpos);
						  }
						  else {
							EMIT2(PUSHLVAL,symtab[$1]->stackpos);
						  }
						  $$=0;
						}
		|
		NARRAY
		'['
		Expression
		']'				{ if (symtab[$1]->status==GLOBAL) {
							EMIT2(GPUSHLVAL,symtab[$1]->stackpos);
						  }
						  else {
							EMIT2(PUSHLVAL,symtab[$1]->stackpos);
						  }
						  $$=1;
						}
		|
		NARRAY
		'['
		Expression
		','
		Expression
		']'				{ if (symtab[$1]->status==GLOBAL) {
							EMIT2(GPUSHLVAL,symtab[$1]->stackpos);
						  }
						  else {
							EMIT2(PUSHLVAL,symtab[$1]->stackpos);
						  }
						  $$=2;
						}
		;

AttrAssignmentStatement:
		AttrSpec
		ASSIGN_TOK
		Expression			{ EMIT1(ATTRSTORE); }
		|
		AttrSpec
		PLUS_ASSIGN_TOK			{ EMIT1(DUPSTACK2);
						  EMIT1(ATTRLOAD); }
						 
		Expression			{ EMIT1(ADD);
						  EMIT1(ATTRSTORE); }
		|
		AttrSpec
		TIMES_ASSIGN_TOK		{ EMIT1(DUPSTACK2);
						  EMIT1(ATTRLOAD); }
		Expression			{ EMIT1(MULTIPLY);
						  EMIT1(ATTRSTORE); }
		;

AttrSpec:
		':'				{ EMIT1(NODELOADTHIS); }
		IDENTIFIER			{ EMIT2(NAMELOAD,$3); }
		|
		NodeExpression
		':'
		IDENTIFIER			{ EMIT2(NAMELOAD,$3); }
		|
		':'				{ EMIT1(NODELOADTHIS); }
		'('
		Expression
		')'
		|
		NodeExpression
		':'
		'('
		Expression
		')'
		;

NodeExpression:
		NodeSpec
		|
		NodeFunctionCall
		NodeCont
		|
		'('
		NodeExpression
		')'
		;

NodeSpec:
		NodeFirst
		NodeCont
		|
		DIVIDE_TOK			{ EMIT1(NODELOADROOT); }
		RootNodeCont
		;

NodeFirst:
		'.'				{ EMIT1(NODELOADTHIS); }
		|
		PARENT_TOK			{ EMIT1(NODELOADTHIS);
						  EMIT1(ASCEND); }
		|
		NIL_TOK				{ EMIT1(NODELOADNIL); }
		|
		NODE				{ EMIT2((symtab[$1]->status==GLOBAL)?GNODELOADVAR:NODELOADVAR,symtab[$1]->stackpos); }
		|
		NARRAY
		'['
		Expression
		']'			{ EMIT2((symtab[$1]->status==GLOBAL)?GVELLOAD1:VELLOAD1,symtab[$1]->stackpos); }
		|
		NARRAY
		'['
		Expression
		','
		Expression
		']'			{ EMIT2((symtab[$1]->status==GLOBAL)?GVELLOAD2:VELLOAD2,symtab[$1]->stackpos); }
		;

NodeCont:
		/* empty */
		|
		DIVIDE_TOK
		NodeContOb
		NodeCont
		;
RootNodeCont:
		/* empty */
		|
		NodeContOb
		NodeCont
		;

NodeContOb:
		IDENTIFIER			{ EMIT2(NODELOAD,$1); }
		NodeSelector			{ if ($3==0)
							EMIT1(DESCEND);
						  else
							EMIT1(DESCENDSEL);
						}
		|
		'('
		Expression
		')'
		NodeSelector			{ if ($4==0)
							EMIT1(DESCEND);
						  else
							EMIT1(DESCENDSEL);
						}
		|
		'.'				{ EMIT1(NODELOADANY); }
		NodeSelector			{ if ($3==0)
							EMIT1(DESCEND);
						  else
							EMIT1(DESCENDSEL);
						}
		|
		PARENT_TOK			{ EMIT1(ASCEND); }
		;

NodeSelector:
		/* empty */			{ $$=0; }
		|
		'['
		Expression
		']'				{ $$=1; }
		;

ProcedureStatement:
		FunctionCall			{ EMIT1(POP); }
		|
		NodeFunctionCall		{ EMIT1(POP); }
		;

IfStatement:
		IF_TOK				{ @1.first_line; 
						  EMIT1(IF); }
		'('
		Expression
		')'				{ EMIT1(CONDEND);
						  EMIT1(JZ);
						  pcpush(codepos); 
						  EMIT1(ERROR); }
		SimpleOrCompoundStatement	{ code[pcpop()] = codepos+2;
						  EMIT1(JMP);
						  pcpush(codepos);
						  EMIT1(ERROR);	}
		ElseStatement			{ code[pcpop()] = codepos; }
		;

ElseStatement:
		/* empty */
		|
		ELSE_TOK
		SimpleOrCompoundStatement
		;

ForStatement:
		FOR_TOK
		'('
		AssignmentStatement
		';'				{ pcpush(codepos); /* stack:top */
						  EMIT1(IF); }
		Expression
		';'				{ EMIT1(CONDEND);
						  EMIT1(JZ);
						  pcpush(codepos); /* stack:top exit-jump */
						  EMIT1(ERROR);
						  EMIT1(JMP);
						  pcpush(codepos); /* stack:top exit-jump block-jump */
						  EMIT1(ERROR);
						}
		AssignmentStatement		{ tmp = pcpop(); 
						  code[tmp] = codepos+2; /* stack:top exit-jump */
						  EMIT1(JMP);
						  pcswap();		/* stack: exit-jump top */
						  EMIT1(pcpop());	/* stack: exit-jump */
						  pcpush(tmp+1);	/* stack: exit-jump assign */
						}
		')'
		SimpleOrCompoundStatement	{ 
						  EMIT1(JMP);
						  EMIT1(pcpop());	/* stack: exit-jump */
						  code[pcpop()] = codepos;
						}
						  
		;

WhileStatement:
		WHILE_TOK
		'('				{ pcpush(codepos); 
						  EMIT1(IF); 
						}
		Expression			{ EMIT1(CONDEND);
						  EMIT1(JZ);
						  pcpush(codepos);
						  EMIT1(ERROR); 
						}
		')'
		SimpleOrCompoundStatement	{ code[pcpop()] = codepos+2;
						  EMIT1(JMP);
						  EMIT1(pcpop());
						}
		;

ForEachStatement:
		FOREACH_TOK
		NODE				{ 
						  if (symtab[$2]->status==GLOBAL)
							EMIT1(GFOREACH);
						  else
							EMIT1(FOREACH);
						  EMIT1(symtab[$2]->stackpos);
						  pcpush(codepos);
						  EMIT1(ERROR);
						  pcpush(codepos);
						  EMIT1(ERROR);
						}
		'('
		NodePattern			{ EMIT1(PATTEND); 
						  code[pcpop()]=codepos; }
		')'
		SimpleOrCompoundStatement	{ EMIT1(FUNCEND); 
						  code[pcpop()]=codepos;
						}
		;

NodePattern:
		NodeFirst
		NodePatternCont			{ EMIT1(PATTSAVE); }
		|
		DIVIDE_TOK			{ EMIT1(NODELOADROOT); }
		RootNodePatternCont		{ EMIT1(PATTSAVE); }
		|
		NodeFunctionCall
		NodePatternCont			{ EMIT1(PATTSAVE); }
		;

NodePatternCont:
		/* empty */
		|
		DIVIDE_TOK
		NodePatternContOb
		NodePatternCont
		;

RootNodePatternCont:
		/* empty */
		|
		NodePatternContOb
		NodePatternCont
		;

NodePatternContOb:
		IDENTIFIER			{ EMIT2(NODELOAD,$1); }
		NodeSelector			{ if ($3==0)
							EMIT1(DESCENDALL);
						  else
							EMIT1(DESCENDSEL);
						}
		|
		'('
		Expression
		')'
		NodeSelector			{ if ($4==0)
							EMIT1(DESCENDALL);
						  else
							EMIT1(DESCENDSEL);
						}
		|
		'.'				{ EMIT1(NODELOADANY); }
		NodeSelector			{ if ($3==0) 
							EMIT1(DESCENDALL);
						  else
							EMIT1(DESCENDSEL);
						}
		|
		PARENT_TOK			{ EMIT1(ASCEND); }
		|
		DESCENDANT_TOK			{ /* EMIT1(DESCENDREC); -- ignore this */ }
		;

ReturnStatement:
		RETURN_TOK
		Expression		{ 
					  if (infunc!=1)
						  if (parsefail("return statement doesn't match function type",&@1)) 
							YYABORT;
					  EMIT1(RETURN);
					  EMIT1(STACKFREE);
					  EMIT1(FUNCEND); 
					}
		|
		RETURN_TOK
		NodeExpression		{ 
					  if (infunc!=2)
						  if (parsefail("return statement doesn't match function type",&@1)) 
							YYABORT;
					  EMIT1(RETURN);
					  EMIT1(STACKFREE);
					  EMIT1(FUNCEND); 
					}
		|
		RETURN_TOK		{
					  if (infunc!=3)
						  if (parsefail("void return statement in element functions only",&@1)) 
							YYABORT;
					  EMIT1(RETURN);
					  EMIT1(STACKFREE);
					  EMIT1(FUNCEND); 
					}
		;

ResizeStatement:
		RESIZE_TOK
		'('
		ARRAY
		','
		Expression
		')'			{ EMIT2((symtab[$3]->status==GLOBAL)?GARRSIZE:ARRSIZE,symtab[$3]->stackpos); }
		|
		RESIZE_TOK
		'('
		NARRAY
		','
		Expression
		')'			{ EMIT2((symtab[$3]->status==GLOBAL)?GARRSIZE:ARRSIZE,symtab[$3]->stackpos); }
		|
		RESIZE_TOK
		'('
		ARRAY
		','
		Expression
		','
		Expression
		')'			{ EMIT2((symtab[$3]->status==GLOBAL)?GARRSIZE2:ARRSIZE2,symtab[$3]->stackpos); }
		|
		RESIZE_TOK
		'('
		NARRAY
		','
		Expression
		','
		Expression
		')'			{ EMIT2((symtab[$3]->status==GLOBAL)?GARRSIZE2:ARRSIZE2,symtab[$3]->stackpos); }
		;

AssertStatement:
		ASSERT_TOK
		'('
		Expression
		')'			{ EMIT1(ASSERT); EMIT1(@1.first_line); }
		|
		ASSERT_TOK
		'('
		Expression
		','
		Expression
		')'			{ EMIT1(ASSERTVAL); EMIT1(@1.first_line); }
		;

RelationalOperator:
		EQ_TOK			{ $$=TESTEQ; }
		|
		NE_TOK			{ $$=TESTNE; }
		|
		LT_TOK			{ $$=TESTLT; }
		|
		LE_TOK			{ $$=TESTLE; }
		|
		GT_TOK			{ $$=TESTGT; }
		|
		GE_TOK			{ $$=TESTGE; }
		;

NodeRelOp:
		EQ_TOK			{ $$=NTESTEQ; }
		|
		NE_TOK			{ $$=NTESTNE; }
		;

Expression:
		SimpleExpression
		;

SimpleExpression:
		Term
		|
		Term
		AddingOperator
		SimpleExpression	{ EMIT1($2); }
		|
		NodeTerm
		|
		NodeTerm
		NodeAddingOperator
		SimpleExpression	{ EMIT1($2); }
		;

AddingOperator:
		PLUS_TOK		{ $$=ADD; }
		|
		MINUS_TOK		{ $$=SUBTRACT; }
		|
		CONCAT_TOK		{ $$=CONCATENATE; }
		|
		OR_TOK			{ $$=LOGICALOR; }
		;

NodeAddingOperator:
		OR_TOK			{ $$=LOGICALOR; }
		;

Term:
		Factor
		|
		Factor
		MultiplyingOperator
		Term			{ EMIT1($2); }
		|
		NodeFactor
		NodeMultiplyingOperator
		Term			{ EMIT1($2); }
		;

NodeTerm:
		NodeFactor
		|
		NodeFactor
		NodeMultiplyingOperator
		NodeTerm			{ EMIT1($2); }
		;

MultiplyingOperator:
		TIMES_TOK		{ $$=MULTIPLY; }
		|
		DIVIDE_TOK		{ $$=DIVIDE; }
		|
		AND_TOK			{ $$=LOGICALAND; }
		;

NodeMultiplyingOperator:
		AND_TOK			{ $$=LOGICALAND; }
		;

Factor:
		FactorComponent
		|
		FactorComponent
		RelationalOperator
		Factor			{ EMIT1($2); }
		;

NodeFactor:
		NodeExpression
		NodeRelOp
		NodeExpression		{ EMIT1($2); }
		;

FactorComponent:
		UnsignedConstant
		|
		'('
		Expression
		')'
		|
		MINUS_TOK
		FactorComponent %prec UNARYMINUS	{ EMIT1(NEGATE); }
		|
		NOT_TOK	
		FactorComponent %prec NOT_TOK	{ EMIT1(LOGICALNOT); }
		|
		STRING			{ EMIT2(CONLOAD,$1); }
		|
		FunctionCall
		|
		SIZEOF_TOK
		'('
		ARRAY
		')'			{ EMIT2((symtab[$3]->status==GLOBAL)?GSIZEOF:SIZEOF,symtab[$3]->stackpos); }
		|
		SIZEOF_TOK
		'('
		NARRAY
		')'			{ EMIT2((symtab[$3]->status==GLOBAL)?GSIZEOF:SIZEOF,symtab[$3]->stackpos); }
		;
		
UnsignedConstant:
		VARIABLE		{ EMIT2((symtab[$1]->status==GLOBAL)?GVARLOAD:VARLOAD,symtab[$1]->stackpos); }
		|
		CONSTSYM		{ EMIT2((symtab[$1]->status==GLOBAL)?GVARLOAD:VARLOAD,symtab[$1]->stackpos); }
		|
		AttrSpec		{ EMIT1(ATTRLOAD); }
		|
		INTEGER			{ EMIT2(CONLOAD,$1); }
		|
		FLOAT			{ EMIT2(CONLOAD,$1); }
 		|
		CHAR			{ EMIT2(CONLOAD,$1); }
		|
		TRUE_TOK		{ EMIT1(LOADTRUE); }
		|
		FALSE_TOK		{ EMIT1(LOADFALSE); }
		|
		ARRAY
		'['
		Expression
		']'			{ EMIT2((symtab[$1]->status==GLOBAL)?GVELLOAD1:VELLOAD1,symtab[$1]->stackpos); }
		|
		ARRAY
		'['
		Expression
		','
		Expression
		']'			{ EMIT2((symtab[$1]->status==GLOBAL)?GVELLOAD2:VELLOAD2,symtab[$1]->stackpos); }
		;

FunctionCall:
		FUNCTION		{ 
					  if (symtab[$1]->status==RESERVED) {
						  if (parsefail("built-in functions cannot be passed as parameters",&@1)) 
							YYABORT;
					  }
					  if (symtab[$1]->status==GLOBAL) {
						  EMIT2(CONLOAD,$1); 
					  }
					  else
						  EMIT2(VARLOAD,symtab[$1]->stackpos); 
					}
		|
		FUNCTION
		'('			{ pcpush(parno); parno=0; }
		ActualParameterList
		')'			{ EMIT2(FUNCTIONCALL,$1);
					  if (parno!=symtab[$1]->nparm) {
						  if (parsefail("incorrect # parameters",&@1)) 
							YYABORT;
					  }
					  parno = pcpop();
					}
		;

NodeFunctionCall:
		NODE_FUNCTION		{ 
					  if (symtab[$1]->status==RESERVED) {
						  if (parsefail("built-in functions cannot be passed as parameters",&@1)) 
							YYABORT;
					  }
					  if (symtab[$1]->status==GLOBAL) {
						  EMIT2(CONLOAD,$1); 
					  }
					  else
						  EMIT2(VARLOAD,symtab[$1]->stackpos); 
					}
		|
		NODE_FUNCTION
		'('			{ pcpush(parno); parno=0; }
		ActualParameterList
		')'			{ EMIT2(FUNCTIONCALL,$1);
					  if (parno!=symtab[$1]->nparm) {
						  if (parsefail("incorrect # parameters",&@1)) 
							YYABORT;
					  }
					  parno = pcpop();
					}
		;

ActualParameterList:
		/* empty */
		|
		ActualParameterListOb
		;

ActualParameterListOb:
		ActualParameter		{ parno++; }
		|
		ActualParameterListOb
		','
		ActualParameter		{ parno++; }
		;

ActualParameter:
		Expression
		|
		NodeExpression
		;

%%

/* local includes */
#include <string.h>

/* global variables */
char	filename[256];		/* script filename */
char	basename[256];		/* source directory */
char	xfilename[256];		/* current XML filename */
char	ofilename[256];		/* output filename */
FILE	*ip;	
int	lno;
int	errcnt;
int	errcomments;

/* program settings */
int	debug=0;		/* produce extra debugging information */
int	doxsave=0;		/* save output back into input file */
int	doanalonly;		/* analyse XML only */
int	domap=0;		/* generate source->PC map */
char	mfilename[256];
FILE	*mfp;

/* main program */
void main(argc,argv)
int	argc;
char	*argv[];
{
	extern char	*optarg;
	extern int	optind;
	int		c,errflg=0;
	char		*p;
	time_t		tim;

	strcpy(ofilename,"-");

	/* scan compiler arguments */
	while ((c=getopt(argc,argv,"IDo:Onm:"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: ProXML Interpreter V%s\n",
					PROGNAME,PROGVERS);
		exit(0);
	case 'o':
		strcpy(ofilename,optarg);
		break;
	case 'O':
		doxsave=1;
		break;
	case 'n':
		doanalonly=1;
		break;
	case 'D':
		debug++;
		break;
	case 'm':		/* map file */
		strcpy(mfilename,optarg);
		domap=1;
		break;
	default:
		errflg=1;
		break;
	}
	if (errflg || (argc < 2))
		prxerror("usage: %s (-I) (-D) (-m mapfile) (-O|-o outfile|-n) program.prx *.xml",PROGNAME);

	/* get program name */
	if (optind < argc)
		strcpy(filename,argv[optind]);
	else
		prxerror("no program file supplied","");
	optind++;

	/* get directory name from file */
	strcpy(basename,filename);
	p = strrchr(basename,'/');
	if (!p) p = strrchr(basename,'\\');
	if (p)
		p[1] = '\0';
	else
		basename[0]='\0';

	/* open program file */
	if ((ip=fopen(filename,"r"))==NULL)
		prxerror("unable to open '%s'",filename);

	/* open map file */
	if (domap) {
		if ((mfp=fopen(mfilename,"w"))==NULL)
			prxerror("unable to open '%s'",mfilename);
		tim = time((time_t *)0);
		fprintf(mfp,"PRX V%s: Compilation of %s on %s",PROGVERS,filename,ctime(&tim));
	}

	/* compile the program */
	if (debug>1) yydebug=1;
	resetlex();
	initsymbol();
	yyparse();
	fclose(ip);
	if (debug) dumpcode();
	if (domap) fclose(mfp);
	if (errcnt!=0) exit(1);

	/* process argument list */
	if (optind==argc) {
		if (doanalonly)
			process(NULL,NULL);
		else
			process(NULL,ofilename);
	}
	else for (;optind<argc;optind++) {
		if (doanalonly)
			process(argv[optind],NULL);
		else if (doxsave)
			process(argv[optind],argv[optind]);
		else
			process(argv[optind],ofilename);
	}

	exit(0);
}

/* error and warning routines */
/* yacc error routine */
void yyerror(s)
char	*s;
{
	fprintf(stderr,"%s(%d) syntax error in line\n",filename,lno);
	errcnt++;
}

int parsefail(s,yylloc)
char	*s;
YYLTYPE	*yylloc;
{
	fprintf(stderr,"%s(%d,%d) %s\n",filename,
				yylloc->first_line,yylloc->first_column,s);
	if (errcnt++ > 20)
		return(1);
	else
		return(0);
}

/* print a warning message */
int warning(yylloc,s,t)
YYLTYPE *yylloc;
char	*s,*t;
{
	fprintf(stderr,"%s(%d,%d) ",filename,
				yylloc->first_line,yylloc->first_column);
	fprintf(stderr,s,t);
	fprintf(stderr,"\n");
	if (errcnt++ > 20)
		return(1);
	else
		return(0);
}

/* print error and die */
void prxerror(s,t)
char	*s,*t;
{
	fprintf(stderr,"%s: ",PROGNAME);
	fprintf(stderr,s,t);
	fprintf(stderr,"\n");
	exit(1);
}
