/* segment list -- SCRIBE broad class annotations */

#define	SIL	0x01	/* silence symbol */
#define NUC	0x02	/* can be a nucleus */
#define SKIP	0x04	/* can be skipped */
#define LONG	0x08	/* has long attribute */
#define STRESS1	0x10	/* primary stress indicator */
#define STRESS2	0x20	/* secondary stree indicator */
#define LINKR	0x40	/* is linking /r/ */
#define ACODA	0x80	/* could go in coda-appendix */
#define BLOCK	0x100	/* block ambi-syllabicity */
#define STOPGAP	0x200	/* stop gap */
#define SYLBOUND 0x400	/* forced syllable boundary */
#define FTBOUND 0x800	/* foot boundary */
#define AGBOUND 0x1000	/* accent group boundary */
#define IPBOUND 0x2000	/* intonation phrase boundary */

struct segment_rec {
	char	*sym;
	int	attr;
	char	*elem;
	char	*elem2;
};

/* table of segment definitions - KEEP SORTED BY sym */
static struct segment_rec segtab[]={
{ " ",	SKIP,		"" },
{ "#",	BLOCK,		"" },
{ "$",	SKIP|SYLBOUND,	"" },
{ "'",	SKIP|STRESS1,	"" },
{ "+",	SKIP,		"" },
{ "+d",	ACODA,		"<CNS CNT='N' VOI='Y' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>d</CNS>", "" },
{ "+s",	ACODA,		"<CNS CNT='Y' VOI='N' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>s</CNS>", "" },
{ "+t",	ACODA,		"<CNS CNT='N' VOI='N' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>t</CNS>", "" },
{ "+z",	ACODA,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>z</CNS>", "" },
{ ",",	SKIP|STRESS2,	"" },
{ "/",	SKIP|IPBOUND,	"" },
{ "3",	NUC|LONG,	"<VOC GRV='Y' HEIGHT='MID' RND='N'>3</VOC>", "<VOC GRV='Y' HEIGHT='MID' RND='N'>3</VOC>" },
{ ":",	SKIP,		"" },
{ "?",	0,		"<CNS CNT='N' VOI='N' NAS='N' STR='N' SON='N' RHO='N'>?</CNS>", "" },
{ "?v",	NUC,		"<VOC GRV='Y' HEIGHT='MID' RND='N'>?</VOC>", "<VOC GRV='Y' HEIGHT='MID' RND='N'>?</VOC>" },
{ "@",	NUC,		"<VOC GRV='Y' HEIGHT='MID' RND='N'>@</VOC>", "<VOC GRV='Y' HEIGHT='MID' RND='N'>@</VOC>" },
{ "@U",	NUC|LONG,	"<VOC GRV='Y' HEIGHT='MID'>@</VOC>", "<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>U</VOC>" },
{ "@U1",NUC|LONG,	"<VOC GRV='Y' HEIGHT='MID'>@</VOC>", "" },
{ "@U2",NUC|LONG,	"<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>U</VOC>", "" },
{ "A",	NUC|LONG,	"<VOC GRV='Y' HEIGHT='OPEN' RND='N'>A</VOC>", "<VOC GRV='Y' HEIGHT='OPEN' RND='N'>A</VOC>" },
{ "D",	0,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>D</CNS>", "" },
{ "I",	NUC,		"<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>", "<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>" },
{ "I@",	NUC|LONG,	"<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>", "<VOC HEIGHT='MID' RND='N'>@</VOC>" },
{ "I@1",NUC|LONG,	"<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>", "" },
{ "I@2",NUC|LONG,	"<VOC HEIGHT='MID' RND='N'>@</VOC>", "" },
{ "N",	0,		"<CNS CNT='N' VOI='Y' NAS='Y' STR='N' SON='Y' RHO='N' CNSGRV='Y' CNSCMP='Y'>N</CNS>", "" },
{ "O",	NUC|LONG,	"<VOC GRV='Y' HEIGHT='MID' RND='Y'>O</VOC>", "<VOC GRV='Y' HEIGHT='MID' RND='Y'>O</VOC>" },
{ "OI",	NUC|LONG,	"<VOC GRV='Y' HEIGHT='MID' RND='Y'>O</VOC>", "<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>" },
{ "OI1",NUC|LONG,	"<VOC GRV='Y' HEIGHT='MID' RND='Y'>O</VOC>", "" },
{ "OI2",NUC|LONG,	"<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>", "" },
{ "Q",	NUC,		"<VOC GRV='Y' HEIGHT='MID' RND='Y'>Q</VOC>", "<VOC GRV='Y' HEIGHT='MID' RND='Y'>Q</VOC>" },
{ "R",	SKIP|LINKR,	"" },
{ "S",	0,		"<CNS CNT='Y' VOI='N' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='Y'>S</CNS>", "" },
{ "T",	0,		"<CNS CNT='Y' VOI='N' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>T</CNS>", "" },
{ "U",	NUC,		"<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>U</VOC>", "<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>U</VOC>" },
{ "U@",	NUC|LONG,	"<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>U</VOC>", "<VOC HEIGHT='MID' RND='N'>@</VOC>" },
{ "U@1",NUC|LONG,	"<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>U</VOC>", "" },
{ "U@2",NUC|LONG,	"<VOC HEIGHT='MID' RND='N'>@</VOC>", "" },
{ "V",	NUC,		"<VOC GRV='Y' HEIGHT='OPEN' RND='N'>V</VOC>", "<VOC GRV='Y' HEIGHT='OPEN' RND='N'>V</VOC>" },
{ "Z",	0,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='Y'>Z</CNS>", "" },
{ "\\",	SKIP|FTBOUND,	"" },
{ "_",	SKIP,		"" },
{ "`",	SKIP|FTBOUND|AGBOUND,	"" },
{ "aI",	NUC|LONG,	"<VOC GRV='Y' HEIGHT='OPEN' RND='N'>a</VOC>", "<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>" },
{ "aI1",NUC|LONG,	"<VOC GRV='Y' HEIGHT='OPEN' RND='N'>a</VOC>", "" },
{ "aI2",NUC|LONG,	"<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>", "" },
{ "aU",	NUC|LONG,	"<VOC GRV='Y' HEIGHT='OPEN' RND='N'>a</VOC>", "<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>U</VOC>" },
{ "aU1",NUC|LONG,	"<VOC GRV='Y' HEIGHT='OPEN' RND='N'>a</VOC>", "" },
{ "aU2",NUC|LONG,	"<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>U</VOC>", "" },
{ "b",	0,		"<CNS CNT='N' VOI='Y' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='Y' CNSCMP='N'>b</CNS>", "" },
{ "b_c",STOPGAP,	"<CNS CNT='N' VOI='Y' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='Y' CNSCMP='N'>b</CNS>", "" },
{ "d",	0,		"<CNS CNT='N' VOI='Y' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>d</CNS>", "" },
{ "dZ",	0,		"<CNS CNT='N' VOI='Y' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='Y'>dZ</CNS>", "" },
{ "dZ_c",STOPGAP,	"<CNS CNT='N' VOI='Y' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='Y'>dZ</CNS>", "" },
{ "d_c",STOPGAP,	"<CNS CNT='N' VOI='Y' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>d</CNS>", "" },
{ "e",	NUC,		"<VOC GRV='N' HEIGHT='MID' RND='N'>e</VOC>", "<VOC GRV='N' HEIGHT='MID' RND='N'>e</VOC>" },
{ "e@",	NUC|LONG,	"<VOC GRV='N' HEIGHT='MID' RND='N'>e</VOC>", "<VOC HEIGHT='MID' RND='N'>@</VOC>" },
{ "e@1",NUC|LONG,	"<VOC GRV='N' HEIGHT='MID' RND='N'>e</VOC>", "" },
{ "e@2",NUC|LONG,	"<VOC HEIGHT='MID' RND='N'>@</VOC>", "" },
{ "eI",	NUC|LONG,	"<VOC GRV='N' HEIGHT='MID' RND='N'>e</VOC>", "<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>" },
{ "eI1",NUC|LONG,	"<VOC GRV='N' HEIGHT='MID' RND='N'>e</VOC>", "" },
{ "eI2",NUC|LONG,	"<VOC GRV='N' HEIGHT='CLOSE' RND='N'>I</VOC>", "" },
{ "f",	0,		"<CNS CNT='Y' VOI='N' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='Y' CNSCMP='N'>f</CNS>", "" },
{ "g",	0,		"<CNS CNT='N' VOI='Y' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='Y' CNSCMP='Y'>g</CNS>", "" },
{ "g_c",STOPGAP,	"<CNS CNT='N' VOI='Y' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='Y' CNSCMP='Y'>g</CNS>", "" },
{ "h",	0,		"<CNS CNT='Y' VOI='N' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='Y' CNSCMP='Y'>h</CNS>", "" },
{ "i",	NUC|LONG,	"<VOC GRV='N' HEIGHT='CLOSE' RND='N'>i</VOC>", "<VOC GRV='N' HEIGHT='CLOSE' RND='N'>i</VOC>" },
{ "j",	0,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='N' SON='Y' RHO='N' CNSGRV='N' CNSCMP='Y' VOCGRV='N' VOCHEIGHT='CLOSE' VOCRND='N'>j</CNS>", "" },
{ "j*",	0,		"<VOC GRV='N' HEIGHT='CLOSE' RND='N'>j</VOC>", "" },
{ "j+",	0,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='N' SON='Y' RHO='N' CNSGRV='N' CNSCMP='Y' VOCGRV='N' VOCHEIGHT='CLOSE' VOCRND='N'>j</CNS>", "" },
{ "ju",	NUC|LONG,	"<VOC GRV='N' HEIGHT='CLOSE' RND='N'>j</VOC>", "<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>u</VOC>" },
{ "k",	0,		"<CNS CNT='N' VOI='N' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='Y' CNSCMP='Y'>k</CNS>", "" },
{ "k_c",STOPGAP,	"<CNS CNT='N' VOI='N' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='Y' CNSCMP='Y'>k</CNS>", "" },
{ "l",	0,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='N' SON='Y' RHO='N' CNSGRV='N' CNSCMP='N' VOCHEIGHT='CLOSE' VOCRND='N'>l</CNS>", "" },
{ "m",	0,		"<CNS CNT='N' VOI='Y' NAS='Y' STR='N' SON='Y' RHO='N' CNSGRV='Y' CNSCMP='N'>m</CNS>", "" },
{ "n",	0,		"<CNS CNT='N' VOI='Y' NAS='Y' STR='N' SON='Y' RHO='N' CNSGRV='N' CNSCMP='N'>n</CNS>", "" },
{ "p",	0,		"<CNS CNT='N' VOI='N' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='Y' CNSCMP='N'>p</CNS>", "" },
{ "p_c",STOPGAP,	"<CNS CNT='N' VOI='N' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='Y' CNSCMP='N'>p</CNS>", "" },
{ "r",	0,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='N' SON='Y' RHO='Y' CNSGRV='N' CNSCMP='N' VOCGRV='Y' VOCHEIGHT='MID' VOCRND='N'>r</CNS>", "" },
{ "s",	0,		"<CNS CNT='Y' VOI='N' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>s</CNS>", "" },
{ "t",	0,		"<CNS CNT='N' VOI='N' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>t</CNS>", "" },
{ "tS",	0,		"<CNS CNT='N' VOI='N' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='Y'>tS</CNS>", "" },
{ "tS_c",STOPGAP,	"<CNS CNT='N' VOI='N' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='Y'>tS</CNS>", "" },
{ "t_c",STOPGAP,	"<CNS CNT='N' VOI='N' NAS='N' STR='N' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>t</CNS>", "" },
{ "u",	NUC|LONG,	"<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>u</VOC>", "<VOC GRV='Y' HEIGHT='CLOSE' RND='Y'>u</VOC>" },
{ "v",	0,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='Y' CNSCMP='N'>v</CNS>", "" },
{ "w",	0,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='N' SON='Y' RHO='N' CNSGRV='Y' VOCGRV='Y' VOCHEIGHT='CLOSE' VOCRND='Y'>w</CNS>", "" },
{ "z",	0,		"<CNS CNT='Y' VOI='Y' NAS='N' STR='Y' SON='N' RHO='N' CNSGRV='N' CNSCMP='N'>z</CNS>", "" },
{ "{",	NUC,		"<VOC GRV='N' HEIGHT='OPEN' RND='N'>{</VOC>", "<VOC GRV='N' HEIGHT='OPEN' RND='N'>{</VOC>" },
};
