/* utt2lab -- XML utterance to ESPS label file */

/* Mark Huckvale - University College London */

/* version 1.0 - September 1998 */

/*
 * utt2lab generates times label files from selected levels in
 * the hierarchical structure
 */
 
#define PROGNAME "utt2lab"
#define PROGVERS "1.0"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"

#ifdef SVR4
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif

/* input file */
char		filename[256];	/* input file */
NSL_Doctype	dct=NULL;
const char	*attrSTART;
const char	*attrSTOP;
char		levelname[256];

/* duplicate a string */
char *strsave(const char *str)
{
	char *ptr=malloc(strlen(str)+1);
	if (!ptr)
		error("out of memory");
	strcpy(ptr,str);
	return(ptr);
}

/* a working version of ObtainItem */
NSL_Item *MyObtainItem(NSL_Data *data,const char *name)
{
	NSL_Item *item;
	for (;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (strcmp(item->label,name)==0)
				return(item);
		}
	}
	return(NULL);
}

/* safe version of atof() */
double myatof(const char *str)
{
	if (str && *str && (isdigit(*str)||(*str=='-')))
		return(atof(str));
	else {
		fprintf(stderr,"?atof(%s)\n",(str)?str:"<null>");
		return(0.0);
	}
}

/* safe version of strcmp */
int mystrcmp(const char *s,const char *t)
{
	if (s && t)
		return(strcmp(s,t));
	else if (s || t)
		return(1);
	else
		return(0);
}

/* format a label entry */
void formlabel(NSL_Item *item,FILE *op,double start,double stop)
{
	static int count=0;
	static double ltime=-1.0;

	if (ltime < 0) fprintf(op,"%.4f 121 /\n",start);
	fprintf(op,"%.4f 121 %s%d\n",stop,item->label,++count);
	ltime = stop;
}

/* create labels from nodes */
void listlabel(NSL_Item *item,FILE *op)
{
	NSL_Data *data;

	if (!item) return;

	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			item = data->first;
			if (levelname[0]) {
				if (stricmp(item->label,levelname)==0) {
					formlabel(item,op,
						myatof(GetAttrStringVal(item,attrSTART)),
						myatof(GetAttrStringVal(item,attrSTOP)));
				}
				else
					listlabel(item,op);
			}
			else {
				if ((stricmp(item->label,"VOC")==0) ||
				    (stricmp(item->label,"CNS")==0)) {
					formlabel(item,op,
						myatof(GetAttrStringVal(item,attrSTART)),
						myatof(GetAttrStringVal(item,attrSTOP)));
				}
				else
					listlabel(item,op);
			}
		}
	}
}

/* process an utterance */
void process(NSL_Item *ipitem,FILE *op)
{
	fprintf(op,"#\n");
	listlabel(ipitem,op);	
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	NSL_File 	inf;
	
	NSL_Item 	*uttitem;
	NSL_Item 	*ipitem;
	NSL_Bit		*bit;
	FILE		*op;

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Il:"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: utterance output as ESPS label file Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 'l':
		strcpy(levelname,optarg);
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-l levelname) input.xml (out.lab)",PROGNAME);

	/* open input file */
	if (optind < argc) {
		strcpy(filename,argv[optind]);
		if ((inf = SFopen(filename, dct, NSL_read))==NULL)
			error("unable to open '%s'",filename);
	}
	else
		error("no input file specified");
	optind++;
	dct = DoctypeFromFile(inf);

	/* get unique names for attributes */
	attrSTART = AttrUniqueName(dct,"START",0);
	attrSTOP = AttrUniqueName(dct,"STOP",0);

	/* read first tag - should be <UTT> */
	if ((bit=GetNextBit(inf))==NULL)
		error("Could not read input file");
	while (bit->type==NSL_pi_bit)
		if ((bit=GetNextBit(inf))==NULL)
			error("Could not read input file");
	if ((bit->type!=NSL_start_bit) || (strcmp(bit->label,"UTT")))
		error("Could not read <UTT> start tag");

	/* load the whole thing into memory */
	uttitem = ItemParse(inf,bit->value.item);

	/* find the IP item */
	ipitem = MyObtainItem(uttitem->data,"IP");

	/* check */
	if (!ipitem)
		error("No IP item in input");

	/* OK, let's go */
	if (optind < argc) {
		if ((op=fopen(argv[optind],"w"))==NULL)
			error("unable to open '%s'",argv[optind]);
		process(ipitem,op);
		fclose(op);
	}
	else
		process(ipitem,stdout);

	/* that's all folks */
	SFclose(inf);
	return 0;
}
