/* uttanal -- take measurements from XML stream and call analysis script */

/* Mark Huckvale - University College London */

/* version 0.1 - July 1998 */

#define PROGNAME "uttanal"
#define PROGVERS "1.0"
char *progname=PROGNAME;

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "lt-memory.h"
#include "nsl.h"
#include "local.h"

/* attributes */
const char *attrID;
const char *attrSTART;
const char *attrSTOP;

/* find first value - breadth first search */
const char *findfirst(NSL_Item *item,const char *attr)
{
	NSL_Item	*nitem;
	NSL_Data	*data;
	const char	*p;

	/* check if exists at top level */
	if ((p=GetAttrStringVal(item,attr))!=NULL)
		return(p);

	/* scan at this level first */
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			nitem = data->first;
			if ((p=GetAttrStringVal(nitem,attr))!=NULL)
				return(p);
		}
	}

	/* scan each lowest level in turn */
	for (data=item->data;data!=NULL;data=data->next) {
		if (data->type == NSL_item_data) {
			nitem = data->first;
			if ((p=findfirst(nitem,attr))!=NULL)
				return(p);
		}
	}

	/* nope */
	return(NULL);
}

/* call script with parameters */
void callscript(char *fname,const char *id,double start,double stop)
{
	char	sysbuf[1024];
	sprintf(sysbuf,"%s %s %g %g",fname,(id)?id:"unknown",start,stop);
	system(sysbuf);
}

/* main program */
int main(int argc, char **argv)
{
	extern int	optind;
	extern char	*optarg;
	int		c,errflg=0;

	char		sfilename[256];
	NSL_Bit		*bit;
	NSL_Item	*uttitem;
	NSL_File 	inf;
	NSL_Doctype 	dct=NULL;
	const char	*pid;
	const char	*p;
	double		start;
	double		stop;

	strcpy(sfilename,"");

	/* Initialise the NSL API */
	NSLInit(0);

	/* option decoding */
	while ((c=getopt(argc,argv,"Is:"))!=EOF) switch (c) {
	case 'I':
		fprintf(stderr,"%s: analyse components by time Vs%s\n",PROGNAME,PROGVERS);
		exit(0);
	case 's':
		strcpy(sfilename,optarg);
		break;
	default:
		errflg=1;
	}
	if (errflg)
		error("usage: %s (-I) (-s script) (list.xml)",PROGNAME);

	/* open input file */
	if (argc==optind)
		inf = SFFopen(stdin, dct, NSL_read, "stdin");
	else if ((inf = SFopen(argv[optind], dct, NSL_read))==NULL)
		error("unable to open '%s'",argv[optind]);
	dct = DoctypeFromFile(inf);


	/* get names of important attributes */
	attrID = AttrUniqueName(dct,"ID",0);
	attrSTART = AttrUniqueName(dct,"START",0);
	attrSTOP = AttrUniqueName(dct,"STOP",0);
	
	/* read first tag - should be <UTT> */
	while ((bit=GetNextBit(inf))!=NULL) {
		if ((bit->type!=NSL_start_bit) || (strcmp(bit->label,"UTT")))
			error("Could not read <UTT> start tag");

		/* load the whole thing into memory */
		uttitem = ItemParse(inf,bit->value.item);

		/* find ID, start and stop */
		pid = GetIdVal(uttitem);
		if (pid==NULL)
			pid = GetAttrStringVal(uttitem,attrID);
		if ((p=findfirst(uttitem,attrSTART))!=NULL)
			start = atof(p);
		else
			start=0;
		if ((p=findfirst(uttitem,attrSTOP))!=NULL)
			stop = atof(p);
		else
			stop=0;

		/* does he want a scipt called */
		if (sfilename[0])
			callscript(sfilename,pid,start,stop);
		else if (pid)
			printf("%s %g %g\n",pid,start,stop);
		else
			printf("%g %g\n",start,stop);
	}
	
	SFclose(inf);
	exit(0);
}
